from __future__ import division, absolute_import
from collections import OrderedDict
import unittest
import os
import sys
import pickle

import numpy as np
from astropy import cosmology
from scipy.stats import ks_2samp
import matplotlib.pyplot as plt
import pandas as pd

import bilby
from bilby.core.prior import Uniform, Constraint
from bilby.gw.prior import BBHPriorDict
from bilby.gw import conversion


class TestBBHPriorDict(unittest.TestCase):

    def setUp(self):
        self.prior_dict = dict()
        self.base_directory =\
            '/'.join(os.path.dirname(
                os.path.abspath(sys.argv[0])).split('/')[:-1])
        self.filename = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                                     'prior_files/binary_black_holes.prior')
        self.bbh_prior_dict = bilby.gw.prior.BBHPriorDict(filename=self.filename)
        for key, value in self.bbh_prior_dict.items():
            self.prior_dict[key] = value

    def tearDown(self):
        del self.prior_dict
        del self.filename
        del self.bbh_prior_dict
        del self.base_directory

    def test_create_default_prior(self):
        default = bilby.gw.prior.BBHPriorDict()
        minima = all([self.bbh_prior_dict[key].minimum == default[key].minimum
                      for key in default.keys()])
        maxima = all([self.bbh_prior_dict[key].maximum == default[key].maximum
                      for key in default.keys()])
        names = all([self.bbh_prior_dict[key].name == default[key].name
                     for key in default.keys()])
        boundaries = all([self.bbh_prior_dict[key].boundary == default[key].boundary
                          for key in default.keys()])

        self.assertTrue(all([minima, maxima, names, boundaries]))

    def test_create_from_dict(self):
        new_dict = bilby.gw.prior.BBHPriorDict(dictionary=self.prior_dict)
        for key in self.bbh_prior_dict:
            self.assertEqual(self.bbh_prior_dict[key], new_dict[key])

    def test_redundant_priors_not_in_dict_before(self):
        for prior in ['chirp_mass', 'total_mass', 'mass_ratio', 'symmetric_mass_ratio',
                      'cos_tilt_1', 'cos_tilt_2', 'phi_1', 'phi_2', 'cos_theta_jn',
                      'comoving_distance', 'redshift']:
            self.assertTrue(self.bbh_prior_dict.test_redundancy(prior))

    def test_redundant_priors_already_in_dict(self):
        for prior in ['mass_1', 'mass_2', 'tilt_1', 'tilt_2',
                      'phi_1', 'phi_2', 'theta_jn', 'luminosity_distance']:
            self.assertTrue(self.bbh_prior_dict.test_redundancy(prior))

    def test_correct_not_redundant_priors_masses(self):
        del self.bbh_prior_dict['mass_2']
        for prior in ['mass_2', 'chirp_mass', 'total_mass',  'symmetric_mass_ratio']:
            self.assertFalse(self.bbh_prior_dict.test_redundancy(prior))

    def test_correct_not_redundant_priors_spin_magnitudes(self):
        del self.bbh_prior_dict['a_2']
        self.assertFalse(self.bbh_prior_dict.test_redundancy('a_2'))

    def test_correct_not_redundant_priors_spin_tilt_1(self):
        del self.bbh_prior_dict['tilt_1']
        for prior in ['tilt_1', 'cos_tilt_1']:
            self.assertFalse(self.bbh_prior_dict.test_redundancy(prior))

    def test_correct_not_redundant_priors_spin_tilt_2(self):
        del self.bbh_prior_dict['tilt_2']
        for prior in ['tilt_2', 'cos_tilt_2']:
            self.assertFalse(self.bbh_prior_dict.test_redundancy(prior))

    def test_correct_not_redundant_priors_spin_azimuth(self):
        del self.bbh_prior_dict['phi_12']
        for prior in ['phi_1', 'phi_2', 'phi_12']:
            self.assertFalse(self.bbh_prior_dict.test_redundancy(prior))

    def test_correct_not_redundant_priors_inclination(self):
        del self.bbh_prior_dict['theta_jn']
        for prior in ['theta_jn', 'cos_theta_jn']:
            self.assertFalse(self.bbh_prior_dict.test_redundancy(prior))

    def test_correct_not_redundant_priors_distance(self):
        del self.bbh_prior_dict['luminosity_distance']
        for prior in ['luminosity_distance', 'comoving_distance',
                      'redshift']:
            self.assertFalse(self.bbh_prior_dict.test_redundancy(prior))

    def test_add_unrelated_prior(self):
        self.assertFalse(self.bbh_prior_dict.test_redundancy('abc'))

    def test_test_has_redundant_priors(self):
        self.assertFalse(self.bbh_prior_dict.test_has_redundant_keys())
        for prior in ['chirp_mass', 'total_mass', 'mass_ratio', 'symmetric_mass_ratio',
                      'cos_tilt_1', 'cos_tilt_2', 'phi_1', 'phi_2', 'cos_theta_jn',
                      'comoving_distance', 'redshift']:
            self.bbh_prior_dict[prior] = 0
            self.assertTrue(self.bbh_prior_dict.test_has_redundant_keys())
            del self.bbh_prior_dict[prior]

    def test_add_constraint_prior_not_redundant(self):
        self.bbh_prior_dict['chirp_mass'] = bilby.prior.Constraint(
            minimum=20, maximum=40, name='chirp_mass')
        self.assertFalse(self.bbh_prior_dict.test_has_redundant_keys())

    def test_pickle_prior(self):
        priors = dict(chirp_mass=bilby.core.prior.Uniform(10, 20),
                      mass_ratio=bilby.core.prior.Uniform(0.125, 1))
        priors = bilby.gw.prior.BBHPriorDict(priors)
        with open("test.pickle", "wb") as file:
            pickle.dump(priors, file)
        with open("test.pickle", "rb") as file:
            priors_loaded = pickle.load(file)
        self.assertEqual(priors, priors_loaded)


class TestPriorConversion(unittest.TestCase):

    def test_bilby_to_lalinference(self):
        mass_1 = [1, 20]
        mass_2 = [1, 20]
        chirp_mass = [1, 5]
        mass_ratio = [0.125, 1]

        bilby_prior = BBHPriorDict(dictionary=dict(
            chirp_mass=Uniform(name='chirp_mass', minimum=chirp_mass[0], maximum=chirp_mass[1]),
            mass_ratio=Uniform(name='mass_ratio', minimum=mass_ratio[0], maximum=mass_ratio[1]),
            mass_2=Constraint(name='mass_2', minimum=mass_1[0], maximum=mass_1[1]),
            mass_1=Constraint(name='mass_1', minimum=mass_2[0], maximum=mass_2[1])))

        lalinf_prior = BBHPriorDict(dictionary=dict(
            mass_ratio=Constraint(name='mass_ratio', minimum=mass_ratio[0], maximum=mass_ratio[1]),
            chirp_mass=Constraint(name='chirp_mass', minimum=chirp_mass[0], maximum=chirp_mass[1]),
            mass_2=Uniform(name='mass_2', minimum=mass_1[0], maximum=mass_1[1]),
            mass_1=Uniform(name='mass_1', minimum=mass_2[0], maximum=mass_2[1])))

        nsamples = 5000
        bilby_samples = bilby_prior.sample(nsamples)
        bilby_samples, _ = conversion.convert_to_lal_binary_black_hole_parameters(
            bilby_samples)

        # Quicker way to generate LA prior samples (rather than specifying Constraint)
        lalinf_samples = []
        while len(lalinf_samples) < nsamples:
            s = lalinf_prior.sample()
            if s["mass_1"] < s["mass_2"]:
                s["mass_1"], s["mass_2"] = s["mass_2"], s["mass_1"]
            if s["mass_2"] / s["mass_1"] > 0.125:
                lalinf_samples.append(s)
        lalinf_samples = pd.DataFrame(lalinf_samples)
        lalinf_samples["mass_ratio"] = lalinf_samples["mass_2"] / lalinf_samples["mass_1"]

        # Construct fake result object
        result = bilby.core.result.Result()
        result.search_parameter_keys = ["mass_ratio", "chirp_mass"]
        result.meta_data = dict()
        result.priors = bilby_prior
        result.posterior = pd.DataFrame(bilby_samples)
        result_converted = bilby.gw.prior.convert_to_flat_in_component_mass_prior(result)

        if "plot" in sys.argv:
            # Useful for debugging
            plt.hist(bilby_samples["mass_ratio"], bins=50, density=True, alpha=0.5)
            plt.hist(result_converted.posterior["mass_ratio"], bins=50, density=True, alpha=0.5)
            plt.hist(lalinf_samples["mass_ratio"], bins=50, alpha=0.5, density=True)
            plt.show()

        # Check that the non-reweighted posteriors fail a KS test
        ks = ks_2samp(bilby_samples["mass_ratio"], lalinf_samples["mass_ratio"])
        print("Non-reweighted KS test = ", ks)
        self.assertFalse(ks.pvalue > 0.05)

        # Check that the non-reweighted posteriors pass a KS test
        ks = ks_2samp(result_converted.posterior["mass_ratio"], lalinf_samples["mass_ratio"])
        print("Reweighted KS test = ", ks)
        self.assertTrue(ks.pvalue > 0.001)


class TestPackagedPriors(unittest.TestCase):
    """ Test that the prepackaged priors load """

    def test_aligned(self):
        filename = 'aligned_spin_binary_black_holes.prior'
        prior_dict = bilby.gw.prior.BBHPriorDict(filename=filename)
        self.assertTrue('chi_1' in prior_dict)
        self.assertTrue('chi_2' in prior_dict)

    def test_precessing(self):
        filename = 'precessing_binary_neutron_stars.prior'
        prior_dict = bilby.gw.prior.BBHPriorDict(filename=filename)
        self.assertTrue('lambda_1' in prior_dict)
        self.assertTrue('lambda_2' in prior_dict)

    def test_binary_black_holes(self):
        filename = 'binary_black_holes.prior'
        prior_dict = bilby.gw.prior.BBHPriorDict(filename=filename)
        self.assertTrue('a_1' in prior_dict)

    def test_binary_neutron_stars(self):
        filename = 'binary_neutron_stars.prior'
        prior_dict = bilby.gw.prior.BNSPriorDict(filename=filename)
        self.assertTrue('lambda_1' in prior_dict)


class TestBNSPriorDict(unittest.TestCase):

    def setUp(self):
        self.prior_dict = OrderedDict()
        self.base_directory =\
            '/'.join(os.path.dirname(
                os.path.abspath(sys.argv[0])).split('/')[:-1])
        self.filename = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                                     'prior_files/binary_neutron_stars.prior')
        self.bns_prior_dict = bilby.gw.prior.BNSPriorDict(filename=self.filename)
        for key, value in self.bns_prior_dict.items():
            self.prior_dict[key] = value

    def tearDown(self):
        del self.prior_dict
        del self.filename
        del self.bns_prior_dict
        del self.base_directory

    def test_create_default_prior(self):
        default = bilby.gw.prior.BNSPriorDict()
        minima = all([self.bns_prior_dict[key].minimum == default[key].minimum
                      for key in default.keys()])
        maxima = all([self.bns_prior_dict[key].maximum == default[key].maximum
                      for key in default.keys()])
        names = all([self.bns_prior_dict[key].name == default[key].name
                     for key in default.keys()])
        boundaries = all([self.bns_prior_dict[key].boundary == default[key].boundary
                          for key in default.keys()])

        self.assertTrue(all([minima, maxima, names, boundaries]))

    def test_create_from_dict(self):
        new_dict = bilby.gw.prior.BNSPriorDict(dictionary=self.prior_dict)
        self.assertDictEqual(self.bns_prior_dict, new_dict)

    def test_redundant_priors_not_in_dict_before(self):
        for prior in ['chirp_mass', 'total_mass', 'mass_ratio',
                      'symmetric_mass_ratio', 'cos_theta_jn', 'comoving_distance',
                      'redshift', 'lambda_tilde', 'delta_lambda_tilde']:
            self.assertTrue(self.bns_prior_dict.test_redundancy(prior))

    def test_redundant_priors_already_in_dict(self):
        for prior in ['mass_1', 'mass_2', 'chi_1', 'chi_2',
                      'theta_jn', 'luminosity_distance',
                      'lambda_1', 'lambda_2']:
            self.assertTrue(self.bns_prior_dict.test_redundancy(prior))

    def test_correct_not_redundant_priors_masses(self):
        del self.bns_prior_dict['mass_2']
        for prior in ['mass_2', 'chirp_mass', 'total_mass',  'symmetric_mass_ratio']:
            self.assertFalse(self.bns_prior_dict.test_redundancy(prior))

    def test_correct_not_redundant_priors_spin_magnitudes(self):
        del self.bns_prior_dict['chi_2']
        self.assertFalse(self.bns_prior_dict.test_redundancy('chi_2'))

    def test_correct_not_redundant_priors_inclination(self):
        del self.bns_prior_dict['theta_jn']
        for prior in ['theta_jn', 'cos_theta_jn']:
            self.assertFalse(self.bns_prior_dict.test_redundancy(prior))

    def test_correct_not_redundant_priors_distance(self):
        del self.bns_prior_dict['luminosity_distance']
        for prior in ['luminosity_distance', 'comoving_distance',
                      'redshift']:
            self.assertFalse(self.bns_prior_dict.test_redundancy(prior))

    def test_correct_not_redundant_priors_tidal(self):
        del self.bns_prior_dict['lambda_1']
        for prior in['lambda_1', 'lambda_tilde', 'delta_lambda_tilde']:
            self.assertFalse(self.bns_prior_dict.test_redundancy(prior))

    def test_add_unrelated_prior(self):
        self.assertFalse(self.bns_prior_dict.test_redundancy('abc'))

    def test_test_has_redundant_priors(self):
        self.assertFalse(self.bns_prior_dict.test_has_redundant_keys())
        for prior in ['chirp_mass', 'total_mass', 'mass_ratio', 'symmetric_mass_ratio',
                      'cos_theta_jn', 'comoving_distance', 'redshift']:
            self.bns_prior_dict[prior] = 0
            self.assertTrue(self.bns_prior_dict.test_has_redundant_keys())
            del self.bns_prior_dict[prior]

    def test_add_constraint_prior_not_redundant(self):
        self.bns_prior_dict['chirp_mass'] = bilby.prior.Constraint(
            minimum=1, maximum=2, name='chirp_mass')
        self.assertFalse(self.bns_prior_dict.test_has_redundant_keys())


class TestCalibrationPrior(unittest.TestCase):

    def setUp(self):
        self.minimum_frequency = 20
        self.maximum_frequency = 1024

    def test_create_constant_uncertainty_spline_prior(self):
        "Test that generated spline prior has the correct number of elements."
        amplitude_sigma = 0.1
        phase_sigma = 0.1
        n_nodes = 9
        label = 'test'
        test = bilby.gw.prior.CalibrationPriorDict.constant_uncertainty_spline(
            amplitude_sigma, phase_sigma, self.minimum_frequency,
            self.maximum_frequency, n_nodes, label)

        self.assertEqual(len(test), n_nodes * 3)


class TestUniformComovingVolumePrior(unittest.TestCase):

    def setUp(self):
        pass

    def test_minimum(self):
        prior = bilby.gw.prior.UniformComovingVolume(
            minimum=10, maximum=10000, name='luminosity_distance')
        self.assertEqual(prior.minimum, 10)

    def test_maximum(self):
        prior = bilby.gw.prior.UniformComovingVolume(
            minimum=10, maximum=10000, name='luminosity_distance')
        self.assertEqual(prior.maximum, 10000)

    def test_zero_minimum_works(self):
        prior = bilby.gw.prior.UniformComovingVolume(
            minimum=0, maximum=10000, name='luminosity_distance')
        self.assertEqual(prior.minimum, 0)

    def test_specify_cosmology(self):
        prior = bilby.gw.prior.UniformComovingVolume(
            minimum=10, maximum=10000, name='luminosity_distance',
            cosmology='Planck13')
        self.assertEqual(repr(prior.cosmology), repr(cosmology.Planck13))

    def test_comoving_prior_creation(self):
        prior = bilby.gw.prior.UniformComovingVolume(
            minimum=10, maximum=1000, name='comoving_distance')
        self.assertEqual(prior.latex_label, '$d_C$')

    def test_redshift_prior_creation(self):
        prior = bilby.gw.prior.UniformComovingVolume(
            minimum=0.1, maximum=1, name='redshift')
        self.assertEqual(prior.latex_label, '$z$')

    def test_redshift_to_luminosity_distance(self):
        prior = bilby.gw.prior.UniformComovingVolume(
            minimum=0.1, maximum=1, name='redshift')
        new_prior = prior.get_corresponding_prior('luminosity_distance')
        self.assertEqual(new_prior.name, 'luminosity_distance')

    def test_luminosity_distance_to_redshift(self):
        prior = bilby.gw.prior.UniformComovingVolume(
            minimum=10, maximum=10000, name='luminosity_distance')
        new_prior = prior.get_corresponding_prior('redshift')
        self.assertEqual(new_prior.name, 'redshift')

    def test_luminosity_distance_to_comoving_distance(self):
        prior = bilby.gw.prior.UniformComovingVolume(
            minimum=10, maximum=10000, name='luminosity_distance')
        new_prior = prior.get_corresponding_prior('comoving_distance')
        self.assertEqual(new_prior.name, 'comoving_distance')


class TestAlignedSpin(unittest.TestCase):

    def setUp(self):
        pass

    def test_default_prior_matches_analytic(self):
        prior = bilby.gw.prior.AlignedSpin()
        chis = np.linspace(-1, 1, 20)
        analytic = - np.log(np.abs(chis)) / 2
        max_difference = max(abs(analytic - prior.prob(chis)))
        self.assertAlmostEqual(max_difference, 0, 2)


if __name__ == '__main__':
    unittest.main()



""" A command line interface for converting resume files into results files """
import argparse
import os
import pickle

import pandas as pd
import bilby as bilby


def setup_command_line_args():
    parser = argparse.ArgumentParser(
        description=__doc__)
    parser.add_argument(
        "resume_files", nargs='+', help="List of resume files")
    parser.add_argument(
        "-f", '--format', default="json", help="Output format, defaults to json",
        choices=["json", "hdf5", "dat"])
    args = parser.parse_args()
    return args


def check_file(resume_file):
    """ Verify the file exists and is a resume file """
    if "resume.pickle" not in resume_file:
        raise ValueError("File {} is not a resume file".format(resume_file))
    if os.path.isfile(resume_file) is False:
        raise ValueError("No file {}".format(resume_file))


def get_outdir_and_label(resume_file):
    """ Infer the appropriate outdir and label from the resume file name """
    label = os.path.basename(resume_file).replace("_resume.pickle", "")
    outdir = os.path.dirname(resume_file)
    return outdir, label


def read_in_pickle_file(resume_file):
    """ Read in the pickle file

    Parameters
    ----------
    resume_file: str
        Input resume file path

    Returns
    -------
    df: pandas.DataFrame
        A data frame of the posterior

    """
    with open(resume_file, "rb") as file:
        data = pickle.load(file)

    if "posterior" in data:
        posterior = data["posterior"]
    else:
        raise ValueError("Resume file has no posterior, unable to convert")

    if "search_parameter_keys" in data:
        search_parameter_keys = data["search_parameter_keys"]
    else:
        search_parameter_keys = ["x{}".format(i) for i in range(posterior.shape[1])]

    df = pd.DataFrame(posterior, columns=search_parameter_keys)
    return df


def convert_df_to_posterior_samples(df, resume_file):
    filename = resume_file.replace("pickle", "dat")
    filename = filename.replace("resume", "preresult")
    df.to_csv(filename, index=False, header=True, sep=' ')


def convert_df_to_preresult(df, format, resume_file):
    outdir, label = get_outdir_and_label(resume_file)
    result = bilby.core.result.Result(
        label=label, outdir=outdir, search_parameter_keys=list(df.keys()))
    result.posterior = df
    result.priors = dict()
    filename = bilby.core.result.result_file_name(outdir, label, format)
    filename = filename.replace("result.{}".format(format), "preresult.{}".format(format))
    result.save_to_file(filename=filename, extension=format)


def convert_resume(resume_file, args):
    check_file(resume_file)
    print("Converting file {} to {}".format(resume_file, args.format))
    df = read_in_pickle_file(resume_file)
    if args.format == "dat":
        convert_df_to_posterior_samples(df, resume_file)
    elif args.format in ["json", "hdf5"]:
        convert_df_to_preresult(df, args.format, resume_file)


def main():
    args = setup_command_line_args()
    for resume_file in args.resume_files:
        convert_resume(resume_file, args)



from __future__ import division, absolute_import
import unittest

from astropy.cosmology import WMAP9, Planck15
from bilby.gw import cosmology


class TestSetCosmology(unittest.TestCase):

    def setUp(self):
        pass

    def test_setting_cosmology_with_string(self):
        cosmology.set_cosmology('WMAP9')
        self.assertEqual(cosmology.COSMOLOGY[1], 'WMAP9')
        cosmology.set_cosmology('Planck15')

    def test_setting_cosmology_with_astropy_object(self):
        cosmology.set_cosmology(WMAP9)
        self.assertEqual(cosmology.COSMOLOGY[1], 'WMAP9')
        cosmology.set_cosmology(Planck15)

    def test_setting_cosmology_with_default(self):
        cosmology.set_cosmology()
        self.assertEqual(cosmology.COSMOLOGY[1], cosmology.DEFAULT_COSMOLOGY.name)

    def test_setting_cosmology_with_flat_lambda_cdm_dict(self):
        cosmo_dict = dict(H0=67.7, Om0=0.3)
        cosmology.set_cosmology(cosmo_dict)
        self.assertEqual(cosmology.COSMOLOGY[1][:13], 'FlatLambdaCDM')

    def test_setting_cosmology_with_lambda_cdm_dict(self):
        cosmo_dict = dict(H0=67.7, Om0=0.3, Ode0=0.7)
        cosmology.set_cosmology(cosmo_dict)
        self.assertEqual(cosmology.COSMOLOGY[1][:9], 'LambdaCDM')

    def test_setting_cosmology_with_w_cdm_dict(self):
        cosmo_dict = dict(H0=67.7, Om0=0.3, Ode0=0.7, w0=-1.0)
        cosmology.set_cosmology(cosmo_dict)
        self.assertEqual(cosmology.COSMOLOGY[1][:4], 'wCDM')


class TestGetCosmology(unittest.TestCase):

    def setUp(self):
        pass

    def test_getting_cosmology_with_string(self):
        self.assertEqual(cosmology.get_cosmology('WMAP9').name, 'WMAP9')

    def test_getting_cosmology_with_default(self):
        self.assertEqual(cosmology.get_cosmology(), cosmology.COSMOLOGY[0])


if __name__ == '__main__':
    unittest.main()



from __future__ import absolute_import, division
import bilby
import unittest
from mock import Mock
import mock
import numpy as np
import os
import scipy.stats as ss


class TestPriorInstantiationWithoutOptionalPriors(unittest.TestCase):

    def setUp(self):
        self.prior = bilby.core.prior.Prior()

    def tearDown(self):
        del self.prior

    def test_name(self):
        self.assertIsNone(self.prior.name)

    def test_latex_label(self):
        self.assertIsNone(self.prior.latex_label)

    def test_is_fixed(self):
        self.assertFalse(self.prior.is_fixed)

    def test_class_instance(self):
        self.assertIsInstance(self.prior, bilby.core.prior.Prior)

    def test_magic_call_is_the_same_as_sampling(self):
        self.prior.sample = Mock(return_value=0.5)
        self.assertEqual(self.prior.sample(), self.prior())

    def test_base_rescale_method(self):
        self.assertIsNone(self.prior.rescale(1))

    def test_base_repr(self):
        """
        We compare that the strings contain all of the same characters in not
        necessarily the same order as python2 doesn't conserve the order of the
        arguments.
        """
        self.prior = bilby.core.prior.Prior(name='test_name', latex_label='test_label', minimum=0, maximum=1,
                                            check_range_nonzero=True, boundary=None)
        expected_string = "Prior(name='test_name', latex_label='test_label', unit=None, minimum=0, maximum=1, " \
                          "check_range_nonzero=True, boundary=None)"
        self.assertTrue(sorted(expected_string) == sorted(self.prior.__repr__()))

    def test_base_prob(self):
        self.assertTrue(np.isnan(self.prior.prob(5)))

    def test_base_ln_prob(self):
        self.prior.prob = lambda val: val
        self.assertEqual(np.log(5), self.prior.ln_prob(5))

    def test_is_in_prior(self):
        self.prior.minimum = 0
        self.prior.maximum = 1
        val_below = self.prior.minimum - 0.1
        val_at_minimum = self.prior.minimum
        val_in_prior = (self.prior.minimum + self.prior.maximum) / 2.
        val_at_maximum = self.prior.maximum
        val_above = self.prior.maximum + 0.1
        self.assertTrue(self.prior.is_in_prior_range(val_at_minimum))
        self.assertTrue(self.prior.is_in_prior_range(val_at_maximum))
        self.assertTrue(self.prior.is_in_prior_range(val_in_prior))
        self.assertFalse(self.prior.is_in_prior_range(val_below))
        self.assertFalse(self.prior.is_in_prior_range(val_above))

    def test_boundary_is_none(self):
        self.assertIsNone(self.prior.boundary)


class TestPriorName(unittest.TestCase):

    def setUp(self):
        self.test_name = 'test_name'
        self.prior = bilby.core.prior.Prior(self.test_name)

    def tearDown(self):
        del self.prior
        del self.test_name

    def test_name_assignment(self):
        self.prior.name = "other_name"
        self.assertEqual(self.prior.name, "other_name")


class TestPriorLatexLabel(unittest.TestCase):
    def setUp(self):
        self.test_name = 'test_name'
        self.prior = bilby.core.prior.Prior(self.test_name)

    def tearDown(self):
        del self.test_name
        del self.prior

    def test_label_assignment(self):
        test_label = 'test_label'
        self.prior.latex_label = 'test_label'
        self.assertEqual(test_label, self.prior.latex_label)

    def test_default_label_assignment(self):
        self.prior.name = 'chirp_mass'
        self.prior.latex_label = None
        self.assertEqual(self.prior.latex_label, '$\mathcal{M}$')

    def test_default_label_assignment_default(self):
        self.assertTrue(self.prior.latex_label, self.prior.name)


class TestPriorIsFixed(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        del self.prior

    def test_is_fixed_parent_class(self):
        self.prior = bilby.core.prior.Prior()
        self.assertFalse(self.prior.is_fixed)

    def test_is_fixed_delta_function_class(self):
        self.prior = bilby.core.prior.DeltaFunction(peak=0)
        self.assertTrue(self.prior.is_fixed)

    def test_is_fixed_uniform_class(self):
        self.prior = bilby.core.prior.Uniform(minimum=0, maximum=10)
        self.assertFalse(self.prior.is_fixed)


class TestPriorBoundary(unittest.TestCase):

    def setUp(self):
        self.prior = bilby.core.prior.Prior(boundary=None)

    def tearDown(self):
        del self.prior

    def test_set_boundary_valid(self):
        self.prior.boundary = 'periodic'
        self.assertEqual(self.prior.boundary, 'periodic')

    def test_set_boundary_invalid(self):
        with self.assertRaises(ValueError):
            self.prior.boundary = 'else'


class TestPriorClasses(unittest.TestCase):

    def setUp(self):

        # set multivariate Gaussian
        mvg = bilby.core.prior.MultivariateGaussianDist(names=['testa', 'testb'],
                                                        mus=[1, 1],
                                                        covs=np.array([[2., 0.5], [0.5, 2.]]),
                                                        weights=1.)
        mvn = bilby.core.prior.MultivariateGaussianDist(names=['testa', 'testb'],
                                                        mus=[1, 1],
                                                        covs=np.array([[2., 0.5], [0.5, 2.]]),
                                                        weights=1.)
        hp_map_file = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                                   'prior_files/GW150914_testing_skymap.fits')
        hp_dist = bilby.gw.prior.HealPixMapPriorDist(hp_map_file,
                                                 names=['testra', 'testdec'])
        hp_3d_dist = bilby.gw.prior.HealPixMapPriorDist(hp_map_file,
                                                        names=['testra', 'testdec', 'testdistance'], distance=True)

        def condition_func(reference_params, test_param):
            return reference_params.copy()

        self.priors = [
            bilby.core.prior.DeltaFunction(name='test', unit='unit', peak=1),
            bilby.core.prior.Gaussian(name='test', unit='unit', mu=0, sigma=1),
            bilby.core.prior.Normal(name='test', unit='unit', mu=0, sigma=1),
            bilby.core.prior.PowerLaw(name='test', unit='unit', alpha=0, minimum=0, maximum=1),
            bilby.core.prior.PowerLaw(name='test', unit='unit', alpha=-1, minimum=0.5, maximum=1),
            bilby.core.prior.PowerLaw(name='test', unit='unit', alpha=2, minimum=1, maximum=1e2),
            bilby.core.prior.Uniform(name='test', unit='unit', minimum=0, maximum=1),
            bilby.core.prior.LogUniform(name='test', unit='unit', minimum=5e0, maximum=1e2),
            bilby.gw.prior.UniformComovingVolume(name='redshift', minimum=0.1, maximum=1.0),
            bilby.gw.prior.UniformSourceFrame(name='redshift', minimum=0.1, maximum=1.0),
            bilby.core.prior.Sine(name='test', unit='unit'),
            bilby.core.prior.Cosine(name='test', unit='unit'),
            bilby.core.prior.Interped(name='test', unit='unit', xx=np.linspace(0, 10, 1000),
                                      yy=np.linspace(0, 10, 1000) ** 4,
                                      minimum=3, maximum=5),
            bilby.core.prior.TruncatedGaussian(name='test', unit='unit', mu=1, sigma=0.4, minimum=-1, maximum=1),
            bilby.core.prior.TruncatedNormal(name='test', unit='unit', mu=1, sigma=0.4, minimum=-1, maximum=1),
            bilby.core.prior.HalfGaussian(name='test', unit='unit', sigma=1),
            bilby.core.prior.HalfNormal(name='test', unit='unit', sigma=1),
            bilby.core.prior.LogGaussian(name='test', unit='unit', mu=0, sigma=1),
            bilby.core.prior.LogNormal(name='test', unit='unit', mu=0, sigma=1),
            bilby.core.prior.Exponential(name='test', unit='unit', mu=1),
            bilby.core.prior.StudentT(name='test', unit='unit', df=3, mu=0, scale=1),
            bilby.core.prior.Beta(name='test', unit='unit', alpha=2.0, beta=2.0),
            bilby.core.prior.Logistic(name='test', unit='unit', mu=0, scale=1),
            bilby.core.prior.Cauchy(name='test', unit='unit', alpha=0, beta=1),
            bilby.core.prior.Lorentzian(name='test', unit='unit', alpha=0, beta=1),
            bilby.core.prior.Gamma(name='test', unit='unit', k=1, theta=1),
            bilby.core.prior.ChiSquared(name='test', unit='unit', nu=2),
            bilby.gw.prior.AlignedSpin(name='test', unit='unit'),
            bilby.core.prior.MultivariateGaussian(dist=mvg, name='testa', unit='unit'),
            bilby.core.prior.MultivariateGaussian(dist=mvg, name='testb', unit='unit'),
            bilby.core.prior.MultivariateNormal(dist=mvn, name='testa', unit='unit'),
            bilby.core.prior.MultivariateNormal(dist=mvn, name='testb', unit='unit'),
            bilby.core.prior.ConditionalDeltaFunction(condition_func=condition_func, name='test', unit='unit', peak=1),
            bilby.core.prior.ConditionalGaussian(condition_func=condition_func, name='test', unit='unit', mu=0, sigma=1),
            bilby.core.prior.ConditionalPowerLaw(condition_func=condition_func, name='test', unit='unit', alpha=0, minimum=0, maximum=1),
            bilby.core.prior.ConditionalPowerLaw(condition_func=condition_func, name='test', unit='unit', alpha=-1, minimum=0.5, maximum=1),
            bilby.core.prior.ConditionalPowerLaw(condition_func=condition_func, name='test', unit='unit', alpha=2, minimum=1, maximum=1e2),
            bilby.core.prior.ConditionalUniform(condition_func=condition_func, name='test', unit='unit', minimum=0, maximum=1),
            bilby.core.prior.ConditionalLogUniform(condition_func=condition_func, name='test', unit='unit', minimum=5e0, maximum=1e2),
            bilby.gw.prior.ConditionalUniformComovingVolume(condition_func=condition_func, name='redshift', minimum=0.1, maximum=1.0),
            bilby.gw.prior.ConditionalUniformSourceFrame(condition_func=condition_func, name='redshift', minimum=0.1, maximum=1.0),
            bilby.core.prior.ConditionalSine(condition_func=condition_func, name='test', unit='unit'),
            bilby.core.prior.ConditionalCosine(condition_func=condition_func, name='test', unit='unit'),
            bilby.core.prior.ConditionalTruncatedGaussian(condition_func=condition_func, name='test', unit='unit', mu=1, sigma=0.4, minimum=-1, maximum=1),
            bilby.core.prior.ConditionalHalfGaussian(condition_func=condition_func, name='test', unit='unit', sigma=1),
            bilby.core.prior.ConditionalLogNormal(condition_func=condition_func, name='test', unit='unit', mu=0, sigma=1),
            bilby.core.prior.ConditionalExponential(condition_func=condition_func, name='test', unit='unit', mu=1),
            bilby.core.prior.ConditionalStudentT(condition_func=condition_func, name='test', unit='unit', df=3, mu=0, scale=1),
            bilby.core.prior.ConditionalBeta(condition_func=condition_func, name='test', unit='unit', alpha=2.0, beta=2.0),
            bilby.core.prior.ConditionalLogistic(condition_func=condition_func, name='test', unit='unit', mu=0, scale=1),
            bilby.core.prior.ConditionalCauchy(condition_func=condition_func, name='test', unit='unit', alpha=0, beta=1),
            bilby.core.prior.ConditionalGamma(condition_func=condition_func, name='test', unit='unit', k=1, theta=1),
            bilby.core.prior.ConditionalChiSquared(condition_func=condition_func, name='test', unit='unit', nu=2),
            bilby.gw.prior.HealPixPrior(dist=hp_dist, name='testra', unit='unit'),
            bilby.gw.prior.HealPixPrior(dist=hp_dist, name='testdec', unit='unit'),
            bilby.gw.prior.HealPixPrior(dist=hp_3d_dist, name='testra', unit='unit'),
            bilby.gw.prior.HealPixPrior(dist=hp_3d_dist, name='testdec', unit='unit'),
            bilby.gw.prior.HealPixPrior(dist=hp_3d_dist, name='testdistance', unit='unit')
        ]

    def tearDown(self):
        del self.priors

    def test_minimum_rescaling(self):
        """Test the the rescaling works as expected."""
        for prior in self.priors:
            if bilby.core.prior.JointPrior in prior.__class__.__mro__:
                minimum_sample = prior.rescale(0)
                if prior.dist.filled_rescale():
                    self.assertAlmostEqual(minimum_sample[0], prior.minimum)
                    self.assertAlmostEqual(minimum_sample[1], prior.minimum)
            else:
                minimum_sample = prior.rescale(0)
                self.assertAlmostEqual(minimum_sample, prior.minimum)

    def test_maximum_rescaling(self):
        """Test the the rescaling works as expected."""
        for prior in self.priors:
            if bilby.core.prior.JointPrior in prior.__class__.__mro__:
                maximum_sample = prior.rescale(0)
                if prior.dist.filled_rescale():
                    self.assertAlmostEqual(maximum_sample[0], prior.maximum)
                    self.assertAlmostEqual(maximum_sample[1], prior.maximum)
            else:
                maximum_sample = prior.rescale(1)
                self.assertAlmostEqual(maximum_sample, prior.maximum)

    def test_many_sample_rescaling(self):
        """Test the the rescaling works as expected."""
        for prior in self.priors:
            many_samples = prior.rescale(np.random.uniform(0, 1, 1000))
            if bilby.core.prior.JointPrior in prior.__class__.__mro__:
                if not prior.dist.filled_rescale():
                    continue
            self.assertTrue(all((many_samples >= prior.minimum) & (many_samples <= prior.maximum)))

    def test_out_of_bounds_rescaling(self):
        """Test the the rescaling works as expected."""
        for prior in self.priors:
            self.assertRaises(ValueError, lambda: prior.rescale(-1))

    def test_least_recently_sampled(self):
        for prior in self.priors:
            least_recently_sampled_expected = prior.sample()
            self.assertEqual(least_recently_sampled_expected, prior.least_recently_sampled)

    def test_sampling_single(self):
        """Test that sampling from the prior always returns values within its domain."""
        for prior in self.priors:
            single_sample = prior.sample()
            self.assertTrue((single_sample >= prior.minimum) & (single_sample <= prior.maximum))

    def test_sampling_many(self):
        """Test that sampling from the prior always returns values within its domain."""
        for prior in self.priors:
            many_samples = prior.sample(5000)
            self.assertTrue((all(many_samples >= prior.minimum)) & (all(many_samples <= prior.maximum)))

    def test_probability_above_domain(self):
        """Test that the prior probability is non-negative in domain of validity and zero outside."""
        for prior in self.priors:
            if prior.maximum != np.inf:
                outside_domain = np.linspace(prior.maximum + 1, prior.maximum + 1e4, 1000)
                if bilby.core.prior.JointPrior in prior.__class__.__mro__:
                    if not prior.dist.filled_request():
                        prior.dist.requested_parameters[prior.name] = outside_domain
                        continue
                self.assertTrue(all(prior.prob(outside_domain) == 0))

    def test_probability_below_domain(self):
        """Test that the prior probability is non-negative in domain of validity and zero outside."""
        for prior in self.priors:
            if prior.minimum != -np.inf:
                outside_domain = np.linspace(prior.minimum - 1e4, prior.minimum - 1, 1000)
                if bilby.core.prior.JointPrior in prior.__class__.__mro__:
                    if not prior.dist.filled_request():
                        prior.dist.requested_parameters[prior.name] = outside_domain
                        continue
                self.assertTrue(all(prior.prob(outside_domain) == 0))

    def test_least_recently_sampled(self):
        for prior in self.priors:
            lrs = prior.sample()
            self.assertEqual(lrs, prior.least_recently_sampled)

    def test_prob_and_ln_prob(self):
        for prior in self.priors:
            sample = prior.sample()
            if not bilby.core.prior.JointPrior in prior.__class__.__mro__:
                # due to the way that the Multivariate Gaussian prior must sequentially call
                # the prob and ln_prob functions, it must be ignored in this test.
                self.assertAlmostEqual(np.log(prior.prob(sample)), prior.ln_prob(sample), 12)

    def test_many_prob_and_many_ln_prob(self):
        for prior in self.priors:
            samples = prior.sample(10)
            if not bilby.core.prior.JointPrior in prior.__class__.__mro__:
                ln_probs = prior.ln_prob(samples)
                probs = prior.prob(samples)
                for sample, logp, p in zip(samples, ln_probs, probs):
                    self.assertAlmostEqual(prior.ln_prob(sample), logp)
                    self.assertAlmostEqual(prior.prob(sample), p)

    def test_cdf_is_inverse_of_rescaling(self):
        domain = np.linspace(0, 1, 100)
        threshold = 1e-9
        for prior in self.priors:
            if isinstance(prior, bilby.core.prior.DeltaFunction) or \
                    bilby.core.prior.JointPrior in prior.__class__.__mro__:
                continue
            rescaled = prior.rescale(domain)
            max_difference = max(np.abs(domain - prior.cdf(rescaled)))
            self.assertLess(max_difference, threshold)

    def test_cdf_one_above_domain(self):
        for prior in self.priors:
            if prior.maximum != np.inf:
                outside_domain = np.linspace(
                    prior.maximum + 1, prior.maximum + 1e4, 1000)
                self.assertTrue(all(prior.cdf(outside_domain) == 1))

    def test_cdf_zero_below_domain(self):
        for prior in self.priors:
            if bilby.core.prior.JointPrior in prior.__class__.__mro__ and prior.maximum == np.inf:
                continue
            if prior.minimum != -np.inf:
                outside_domain = np.linspace(
                    prior.minimum - 1e4, prior.minimum - 1, 1000)
                self.assertTrue(all(
                    np.nan_to_num(prior.cdf(outside_domain)) == 0))

    def test_log_normal_fail(self):
        with self.assertRaises(ValueError):
            bilby.core.prior.LogNormal(name='test', unit='unit', mu=0, sigma=-1)

    def test_studentt_fail(self):
        with self.assertRaises(ValueError):
            bilby.core.prior.StudentT(name='test', unit='unit', df=3, mu=0, scale=-1)
        with self.assertRaises(ValueError):
            bilby.core.prior.StudentT(name='test', unit='unit', df=0, mu=0, scale=1)

    def test_beta_fail(self):
        with self.assertRaises(ValueError):
            bilby.core.prior.Beta(name='test', unit='unit', alpha=-2.0, beta=2.0),

        with self.assertRaises(ValueError):
            bilby.core.prior.Beta(name='test', unit='unit', alpha=2.0, beta=-2.0),

    def test_multivariate_gaussian_fail(self):
        with self.assertRaises(ValueError):
            # bounds is wrong length
            bilby.core.prior.MultivariateGaussianDist(['a', 'b'],
                                                      bounds=[(-1., 1.)])
        with self.assertRaises(ValueError):
            # bounds has lower value greater than upper
            bilby.core.prior.MultivariateGaussianDist(['a', 'b'],
                                                      bounds=[(-1., 1.), (1., -1)])
        with self.assertRaises(TypeError):
            # bound is not a list/tuple
            bilby.core.prior.MultivariateGaussianDist(['a', 'b'],
                                                      bounds=[(-1., 1.), 2])
        with self.assertRaises(ValueError):
            # bound contains too many values
            bilby.core.prior.MultivariateGaussianDist(['a', 'b'],
                                                      bounds=[(-1., 1., 4), 2])
        with self.assertRaises(ValueError):
            # means is not a list
            bilby.core.prior.MultivariateGaussianDist(['a', 'b'], mus=1.)
        with self.assertRaises(ValueError):
            # sigmas is not a list
            bilby.core.prior.MultivariateGaussianDist(['a', 'b'], sigmas=1.)
        with self.assertRaises(TypeError):
            # covariances is not a list
            bilby.core.prior.MultivariateGaussianDist(['a', 'b'], covs=1.)
        with self.assertRaises(TypeError):
            # correlation coefficients is not a list
            bilby.core.prior.MultivariateGaussianDist(['a', 'b'], corrcoefs=1.)
        with self.assertRaises(ValueError):
            # wrong number of weights
            bilby.core.prior.MultivariateGaussianDist(['a', 'b'], weights=[0.5, 0.5])
        with self.assertRaises(ValueError):
            # not enough modes set
            bilby.core.prior.MultivariateGaussianDist(['a', 'b'], mus=[[1., 2.]],
                                                      nmodes=2)
        with self.assertRaises(ValueError):
            # covariance is the wrong shape
            bilby.core.prior.MultivariateGaussianDist(['a', 'b'],
                                                      covs=np.array([[[1., 1.],
                                                                      [1., 1.]]]))
        with self.assertRaises(ValueError):
            # covariance is the wrong shape
            bilby.core.prior.MultivariateGaussianDist(['a', 'b'],
                                                      covs=np.array([[[1., 1.]]]))
        with self.assertRaises(ValueError):
            # correlation coefficient matrix is the wrong shape
            bilby.core.prior.MultivariateGaussianDist(['a', 'b'], sigmas=[1., 1.],
                                                      corrcoefs=np.array([[[[1., 1.],
                                                                            [1., 1.]]]]))
        with self.assertRaises(ValueError):
            # correlation coefficient matrix is the wrong shape
            bilby.core.prior.MultivariateGaussianDist(['a', 'b'], sigmas=[1., 1.],
                                                      corrcoefs=np.array([[[1., 1.]]]))
        with self.assertRaises(ValueError):
            # correlation coefficient has non-unity diagonal value
            bilby.core.prior.MultivariateGaussianDist(['a', 'b'], sigmas=[1., 1.],
                                                      corrcoefs=np.array([[1., 1.],
                                                                          [1., 2.]]))
        with self.assertRaises(ValueError):
            # correlation coefficient matrix is not symmetric
            bilby.core.prior.MultivariateGaussianDist(['a', 'b'], sigmas=[1., 2.],
                                                      corrcoefs=np.array([[1., -1.2],
                                                                          [-0.3, 1.]]))
        with self.assertRaises(ValueError):
            # correlation coefficient matrix is not positive definite
            bilby.core.prior.MultivariateGaussianDist(['a', 'b'], sigmas=[1., 2.],
                                                      corrcoefs=np.array([[1., -1.3],
                                                                          [-1.3, 1.]]))
        with self.assertRaises(ValueError):
            # wrong number of sigmas
            bilby.core.prior.MultivariateGaussianDist(['a', 'b'], sigmas=[1., 2., 3.],
                                                      corrcoefs=np.array([[1., 0.3],
                                                                          [0.3, 1.]]))

    def test_multivariate_gaussian_covariance(self):
        """Test that the correlation coefficient/covariance matrices are correct"""
        cov = np.array([[4., 0], [0., 9.]])
        mvg = bilby.core.prior.MultivariateGaussianDist(['a', 'b'], covs=cov)
        self.assertEqual(mvg.nmodes, 1)
        self.assertTrue(np.allclose(mvg.covs[0], cov))
        self.assertTrue(np.allclose(mvg.sigmas[0], np.sqrt(np.diag(cov))))
        self.assertTrue(np.allclose(mvg.corrcoefs[0], np.eye(2)))

        corrcoef = np.array([[1., 0.5], [0.5, 1.]])
        sigma = [2., 2.]
        mvg = bilby.core.prior.MultivariateGaussianDist(['a', 'b'],
                                                        corrcoefs=corrcoef,
                                                        sigmas=sigma)
        self.assertTrue(np.allclose(mvg.corrcoefs[0], corrcoef))
        self.assertTrue(np.allclose(mvg.sigmas[0], sigma))
        self.assertTrue(np.allclose(np.diag(mvg.covs[0]), np.square(sigma)))
        self.assertTrue(np.allclose(np.diag(np.fliplr(mvg.covs[0])), 2.*np.ones(2)))

    def test_fermidirac_fail(self):
        with self.assertRaises(ValueError):
            bilby.core.prior.FermiDirac(name='test', unit='unit', sigma=1.)

        with self.assertRaises(ValueError):
            bilby.core.prior.FermiDirac(name='test', unit='unit', sigma=1., mu=-1)

    def test_probability_in_domain(self):
        """Test that the prior probability is non-negative in domain of validity and zero outside."""
        for prior in self.priors:
            if prior.minimum == -np.inf:
                prior.minimum = -1e5
            if prior.maximum == np.inf:
                prior.maximum = 1e5
            domain = np.linspace(prior.minimum, prior.maximum, 1000)
            self.assertTrue(all(prior.prob(domain) >= 0))

    def test_probability_surrounding_domain(self):
        """Test that the prior probability is non-negative in domain of validity and zero outside."""
        for prior in self.priors:
            # skip delta function prior in this case
            if isinstance(prior, bilby.core.prior.DeltaFunction):
                continue
            surround_domain = np.linspace(prior.minimum - 1, prior.maximum + 1, 1000)
            indomain = (surround_domain >= prior.minimum) | (surround_domain <= prior.maximum)
            outdomain = (surround_domain < prior.minimum) | (surround_domain > prior.maximum)
            if bilby.core.prior.JointPrior in prior.__class__.__mro__:
                if not prior.dist.filled_request():
                    continue
            self.assertTrue(all(prior.prob(surround_domain[indomain]) >= 0))
            self.assertTrue(all(prior.prob(surround_domain[outdomain]) == 0))

    def test_normalized(self):
        """Test that each of the priors are normalised, this needs care for delta function and Gaussian priors"""
        for prior in self.priors:
            if isinstance(prior, bilby.core.prior.DeltaFunction):
                continue
            if isinstance(prior, bilby.core.prior.Cauchy):
                continue
            if bilby.core.prior.JointPrior in prior.__class__.__mro__:
                continue
            elif isinstance(prior, bilby.core.prior.Gaussian):
                domain = np.linspace(-1e2, 1e2, 1000)
            elif isinstance(prior, bilby.core.prior.Cauchy):
                domain = np.linspace(-1e2, 1e2, 1000)
            elif isinstance(prior, bilby.core.prior.StudentT):
                domain = np.linspace(-1e2, 1e2, 1000)
            elif isinstance(prior, bilby.core.prior.HalfGaussian):
                domain = np.linspace(0., 1e2, 1000)
            elif isinstance(prior, bilby.core.prior.Gamma):
                domain = np.linspace(0., 1e2, 5000)
            elif isinstance(prior, bilby.core.prior.LogNormal):
                domain = np.linspace(0., 1e2, 1000)
            elif isinstance(prior, bilby.core.prior.Exponential):
                domain = np.linspace(0., 1e2, 5000)
            elif isinstance(prior, bilby.core.prior.Logistic):
                domain = np.linspace(-1e2, 1e2, 1000)
            elif isinstance(prior, bilby.core.prior.FermiDirac):
                domain = np.linspace(0., 1e2, 1000)
            else:
                domain = np.linspace(prior.minimum, prior.maximum, 1000)
            self.assertAlmostEqual(np.trapz(prior.prob(domain), domain), 1, 3)

    def test_accuracy(self):
        """Test that each of the priors' functions is calculated accurately, as compared to scipy's calculations"""
        for prior in self.priors:
            rescale_domain = np.linspace(0, 1, 1000)
            if isinstance(prior, bilby.core.prior.Uniform):
                domain = np.linspace(-5, 5, 100)
                scipy_prob = ss.uniform.pdf(domain, loc=0, scale=1)
                scipy_lnprob = ss.uniform.logpdf(domain, loc=0, scale=1)
                scipy_cdf = ss.uniform.cdf(domain, loc=0, scale=1)
                scipy_rescale = ss.uniform.ppf(rescale_domain, loc=0, scale=1)
            elif isinstance(prior, bilby.core.prior.Gaussian):
                domain = np.linspace(-1e2, 1e2, 1000)
                scipy_prob = ss.norm.pdf(domain, loc=0, scale=1)
                scipy_lnprob = ss.norm.logpdf(domain, loc=0, scale=1)
                scipy_cdf = ss.norm.cdf(domain, loc=0, scale=1)
                scipy_rescale = ss.norm.ppf(rescale_domain, loc=0, scale=1)
            elif isinstance(prior, bilby.core.prior.Cauchy):
                domain = np.linspace(-1e2, 1e2, 1000)
                scipy_prob = ss.cauchy.pdf(domain, loc=0, scale=1)
                scipy_lnprob = ss.cauchy.logpdf(domain, loc=0, scale=1)
                scipy_cdf = ss.cauchy.cdf(domain, loc=0, scale=1)
                scipy_rescale = ss.cauchy.ppf(rescale_domain, loc=0, scale=1)
            elif isinstance(prior, bilby.core.prior.StudentT):
                domain = np.linspace(-1e2, 1e2, 1000)
                scipy_prob = ss.t.pdf(domain, 3, loc=0, scale=1)
                scipy_lnprob = ss.t.logpdf(domain, 3, loc=0, scale=1)
                scipy_cdf = ss.t.cdf(domain, 3, loc=0, scale=1)
                scipy_rescale = ss.t.ppf(rescale_domain, 3, loc=0, scale=1)
            elif (isinstance(prior, bilby.core.prior.Gamma) and
                    not isinstance(prior, bilby.core.prior.ChiSquared)):
                domain = np.linspace(0., 1e2, 5000)
                scipy_prob = ss.gamma.pdf(domain, 1, loc=0, scale=1)
                scipy_lnprob = ss.gamma.logpdf(domain, 1, loc=0, scale=1)
                scipy_cdf = ss.gamma.cdf(domain, 1, loc=0, scale=1)
                scipy_rescale = ss.gamma.ppf(rescale_domain, 1, loc=0, scale=1)
            elif isinstance(prior, bilby.core.prior.LogNormal):
                domain = np.linspace(0., 1e2, 1000)
                scipy_prob = ss.lognorm.pdf(domain, 1, scale=1)
                scipy_lnprob = ss.lognorm.logpdf(domain, 1, scale=1)
                scipy_cdf = ss.lognorm.cdf(domain, 1, scale=1)
                scipy_rescale = ss.lognorm.ppf(rescale_domain, 1, scale=1)
            elif isinstance(prior, bilby.core.prior.Exponential):
                domain = np.linspace(0., 1e2, 5000)
                scipy_prob = ss.expon.pdf(domain, scale=1)
                scipy_lnprob = ss.expon.logpdf(domain, scale=1)
                scipy_cdf = ss.expon.cdf(domain, scale=1)
                scipy_rescale = ss.expon.ppf(rescale_domain, scale=1)
            elif isinstance(prior, bilby.core.prior.Logistic):
                domain = np.linspace(-1e2, 1e2, 1000)
                scipy_prob = ss.logistic.pdf(domain, loc=0, scale=1)
                scipy_lnprob = ss.logistic.logpdf(domain, loc=0, scale=1)
                scipy_cdf = ss.logistic.cdf(domain, loc=0, scale=1)
                scipy_rescale = ss.logistic.ppf(rescale_domain, loc=0, scale=1)
            elif isinstance(prior, bilby.core.prior.ChiSquared):
                domain = np.linspace(0., 1e2, 5000)
                scipy_prob = ss.gamma.pdf(domain, 1, loc=0, scale=2)
                scipy_lnprob = ss.gamma.logpdf(domain, 1, loc=0, scale=2)
                scipy_cdf = ss.gamma.cdf(domain, 1, loc=0, scale=2)
                scipy_rescale = ss.gamma.ppf(rescale_domain, 1, loc=0, scale=2)
            elif isinstance(prior, bilby.core.prior.Beta):
                domain = np.linspace(-5, 5, 5000)
                scipy_prob = ss.beta.pdf(domain, 2, 2, loc=0, scale=1)
                scipy_lnprob = ss.beta.logpdf(domain, 2, 2, loc=0, scale=1)
                scipy_cdf = ss.beta.cdf(domain, 2, 2, loc=0, scale=1)
                scipy_rescale = ss.beta.ppf(rescale_domain, 2, 2, loc=0, scale=1)
            else:
                continue
            testTuple = (
                bilby.core.prior.Uniform, bilby.core.prior.Gaussian,
                bilby.core.prior.Cauchy, bilby.core.prior.StudentT,
                bilby.core.prior.Exponential, bilby.core.prior.Logistic,
                bilby.core.prior.LogNormal, bilby.core.prior.Gamma,
                bilby.core.prior.Beta)
            if isinstance(prior, (testTuple)):
                np.testing.assert_almost_equal(prior.prob(domain), scipy_prob)
                np.testing.assert_almost_equal(prior.ln_prob(domain), scipy_lnprob)
                np.testing.assert_almost_equal(prior.cdf(domain), scipy_cdf)
                np.testing.assert_almost_equal(prior.rescale(rescale_domain), scipy_rescale)

    def test_unit_setting(self):
        for prior in self.priors:
            if isinstance(prior, bilby.gw.prior.Cosmological):
                self.assertEqual(None, prior.unit)
            else:
                self.assertEqual('unit', prior.unit)

    def test_eq_different_classes(self):
        for i in range(len(self.priors)):
            for j in range(len(self.priors)):
                if i == j:
                    self.assertEqual(self.priors[i], self.priors[j])
                else:
                    self.assertNotEqual(self.priors[i], self.priors[j])

    def test_eq_other_condition(self):
        prior_1 = bilby.core.prior.PowerLaw(name='test', unit='unit', alpha=0, minimum=0, maximum=1)
        prior_2 = bilby.core.prior.PowerLaw(name='test', unit='unit', alpha=0, minimum=0, maximum=1.5)
        self.assertNotEqual(prior_1, prior_2)

    def test_eq_different_keys(self):
        prior_1 = bilby.core.prior.PowerLaw(name='test', unit='unit', alpha=0, minimum=0, maximum=1)
        prior_2 = bilby.core.prior.PowerLaw(name='test', unit='unit', alpha=0, minimum=0, maximum=1)
        prior_2.other_key = 5
        self.assertNotEqual(prior_1, prior_2)

    def test_np_array_eq(self):
        prior_1 = bilby.core.prior.PowerLaw(name='test', unit='unit', alpha=0, minimum=0, maximum=1)
        prior_2 = bilby.core.prior.PowerLaw(name='test', unit='unit', alpha=0, minimum=0, maximum=1)
        prior_1.array_attribute = np.array([1, 2, 3])
        prior_2.array_attribute = np.array([2, 2, 3])
        self.assertNotEqual(prior_1, prior_2)

    def test_repr(self):
        for prior in self.priors:
            if isinstance(prior, bilby.core.prior.Interped):
                continue  # we cannot test this because of the numpy arrays
            elif isinstance(prior, bilby.core.prior.MultivariateGaussian):
                repr_prior_string = 'bilby.core.prior.' + repr(prior)
                repr_prior_string = repr_prior_string.replace(
                    'MultivariateGaussianDist',
                    'bilby.core.prior.MultivariateGaussianDist'
                )
            elif isinstance(prior, bilby.gw.prior.HealPixPrior):
                repr_prior_string = 'bilby.gw.prior.'+repr(prior)
                repr_prior_string = repr_prior_string.replace('HealPixMapPriorDist',
                                                              'bilby.gw.prior.HealPixMapPriorDist')
            elif isinstance(prior, bilby.gw.prior.UniformComovingVolume):
                repr_prior_string = 'bilby.gw.prior.' + repr(prior)
            elif 'Conditional' in prior.__class__.__name__:
                continue # This feature does not exist because we cannot recreate the condition function
            else:
                repr_prior_string = 'bilby.core.prior.' + repr(prior)
            repr_prior = eval(repr_prior_string, None, dict(inf=np.inf))
            self.assertEqual(prior, repr_prior)

    def test_set_maximum_setting(self):
        for prior in self.priors:
            if isinstance(prior, (
                    bilby.core.prior.DeltaFunction, bilby.core.prior.Gaussian,
                    bilby.core.prior.HalfGaussian, bilby.core.prior.LogNormal,
                    bilby.core.prior.Exponential, bilby.core.prior.StudentT,
                    bilby.core.prior.Logistic, bilby.core.prior.Cauchy,
                    bilby.core.prior.Gamma, bilby.core.prior.MultivariateGaussian,
                    bilby.core.prior.FermiDirac, bilby.gw.prior.HealPixPrior)):
                continue
            prior.maximum = (prior.maximum + prior.minimum) / 2
            self.assertTrue(max(prior.sample(10000)) < prior.maximum)

    def test_set_minimum_setting(self):
        for prior in self.priors:
            if isinstance(prior, (
                    bilby.core.prior.DeltaFunction, bilby.core.prior.Gaussian,
                    bilby.core.prior.HalfGaussian, bilby.core.prior.LogNormal,
                    bilby.core.prior.Exponential, bilby.core.prior.StudentT,
                    bilby.core.prior.Logistic, bilby.core.prior.Cauchy,
                    bilby.core.prior.Gamma, bilby.core.prior.MultivariateGaussian,
                    bilby.core.prior.FermiDirac, bilby.gw.prior.HealPixPrior)):
                continue
            prior.minimum = (prior.maximum + prior.minimum) / 2
            self.assertTrue(min(prior.sample(10000)) > prior.minimum)


class TestPriorDict(unittest.TestCase):

    def setUp(self):
        self.first_prior = bilby.core.prior.Uniform(name='a', minimum=0, maximum=1, unit='kg', boundary=None)
        self.second_prior = bilby.core.prior.PowerLaw(name='b', alpha=3, minimum=1, maximum=2, unit='m/s',
                                                      boundary=None)
        self.third_prior = bilby.core.prior.DeltaFunction(name='c', peak=42, unit='m')
        self.priors = dict(mass=self.first_prior,
                           speed=self.second_prior,
                           length=self.third_prior)
        self.prior_set_from_dict = bilby.core.prior.PriorDict(dictionary=self.priors)
        self.default_prior_file = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                                               'prior_files/binary_black_holes.prior')
        self.prior_set_from_file = bilby.core.prior.PriorDict(filename=self.default_prior_file)

    def tearDown(self):
        del self.first_prior
        del self.second_prior
        del self.third_prior
        del self.priors
        del self.prior_set_from_dict
        del self.default_prior_file
        del self.prior_set_from_file

    def test_copy(self):
        priors = bilby.core.prior.PriorDict(self.priors)
        self.assertEqual(priors, priors.copy())

    def test_prior_set(self):
        priors_dict = bilby.core.prior.PriorDict(self.priors)
        priors_set = bilby.core.prior.PriorSet(self.priors)
        self.assertEqual(priors_dict, priors_set)

    def test_prior_set_is_dict(self):
        self.assertIsInstance(self.prior_set_from_dict, dict)

    def test_prior_set_has_correct_length(self):
        self.assertEqual(3, len(self.prior_set_from_dict))

    def test_prior_set_has_expected_priors(self):
        self.assertDictEqual(self.priors, dict(self.prior_set_from_dict))

    def test_read_from_file(self):
        expected = dict(
            mass_1=bilby.core.prior.Uniform(
                name='mass_1', minimum=5, maximum=100, unit='$M_{\\odot}$', boundary=None),
            mass_2=bilby.core.prior.Uniform(
                name='mass_2', minimum=5, maximum=100, unit='$M_{\\odot}$', boundary=None),
            mass_ratio=bilby.core.prior.Constraint(name='mass_ratio', minimum=0.125, maximum=1, latex_label='$q$',
                                                   unit=None),
            a_1=bilby.core.prior.Uniform(name='a_1', minimum=0, maximum=0.8, boundary='reflective'),
            a_2=bilby.core.prior.Uniform(name='a_2', minimum=0, maximum=0.8, boundary='reflective'),
            tilt_1=bilby.core.prior.Sine(name='tilt_1', boundary='reflective'),
            tilt_2=bilby.core.prior.Sine(name='tilt_2', boundary='reflective'),
            phi_12=bilby.core.prior.Uniform(
                name='phi_12', minimum=0, maximum=2 * np.pi, boundary='periodic'),
            phi_jl=bilby.core.prior.Uniform(
                name='phi_jl', minimum=0, maximum=2 * np.pi, boundary='periodic'),
            luminosity_distance=bilby.gw.prior.UniformSourceFrame(
                name='luminosity_distance', minimum=1e2,
                maximum=5e3, unit='Mpc', boundary=None),
            dec=bilby.core.prior.Cosine(name='dec', boundary='reflective'),
            ra=bilby.core.prior.Uniform(
                name='ra', minimum=0, maximum=2 * np.pi, boundary='periodic'),
            theta_jn=bilby.core.prior.Sine(name='theta_jn', boundary='reflective'),
            psi=bilby.core.prior.Uniform(name='psi', minimum=0, maximum=np.pi, boundary='periodic'),
            phase=bilby.core.prior.Uniform(
                name='phase', minimum=0, maximum=2 * np.pi, boundary='periodic')
            )
        self.assertDictEqual(expected, self.prior_set_from_file)

    def test_to_file(self):
        """
        We compare that the strings contain all of the same characters in not
        necessarily the same order as python2 doesn't conserve the order of the
        arguments.
        """
        expected = ["length = DeltaFunction(peak=42, name='c', latex_label='c', unit='m')\n",
                    "speed = PowerLaw(alpha=3, minimum=1, maximum=2, name='b', latex_label='b', "
                    "unit='m/s', boundary=None)\n",
                    "mass = Uniform(minimum=0, maximum=1, name='a', latex_label='a', "
                    "unit='kg', boundary=None)\n"]
        self.prior_set_from_dict.to_file(outdir='prior_files', label='to_file_test')
        with open('prior_files/to_file_test.prior') as f:
            for i, line in enumerate(f.readlines()):
                self.assertTrue(any([
                    sorted(line) == sorted(expect) for expect in expected]))

    def test_from_dict_with_string(self):
        string_prior = "PowerLaw(name='b', alpha=3, minimum=1, maximum=2, unit='m/s', " \
                       "boundary=None)"
        self.priors['speed'] = string_prior
        from_dict = bilby.core.prior.PriorDict(dictionary=self.priors)
        self.assertDictEqual(self.prior_set_from_dict, from_dict)

    def test_convert_floats_to_delta_functions(self):
        self.prior_set_from_dict['d'] = 5
        self.prior_set_from_dict['e'] = 7.3
        self.prior_set_from_dict['f'] = 'unconvertable'
        self.prior_set_from_dict.convert_floats_to_delta_functions()
        expected = dict(mass=bilby.core.prior.Uniform(name='a', minimum=0, maximum=1, unit='kg',
                                                      boundary=None),
                        speed=bilby.core.prior.PowerLaw(name='b', alpha=3, minimum=1, maximum=2, unit='m/s',
                                                        boundary=None),
                        length=bilby.core.prior.DeltaFunction(name='c', peak=42, unit='m'),
                        d=bilby.core.prior.DeltaFunction(peak=5),
                        e=bilby.core.prior.DeltaFunction(peak=7.3),
                        f='unconvertable')
        self.assertDictEqual(expected, self.prior_set_from_dict)

    def test_prior_set_from_dict_but_using_a_string(self):
        prior_set = bilby.core.prior.PriorDict(dictionary=self.default_prior_file)
        expected = bilby.core.prior.PriorDict(
            dict(
                mass_1=bilby.core.prior.Uniform(
                    name='mass_1', minimum=5, maximum=100, unit='$M_{\\odot}$', boundary=None),
                mass_2=bilby.core.prior.Uniform(
                    name='mass_2', minimum=5, maximum=100, unit='$M_{\\odot}$', boundary=None),
                mass_ratio=bilby.core.prior.Constraint(name='mass_ratio', minimum=0.125, maximum=1, latex_label='$q$',
                                                       unit=None),
                a_1=bilby.core.prior.Uniform(name='a_1', minimum=0, maximum=0.8, boundary='reflective'),
                a_2=bilby.core.prior.Uniform(name='a_2', minimum=0, maximum=0.8, boundary='reflective'),
                tilt_1=bilby.core.prior.Sine(name='tilt_1', boundary='reflective'),
                tilt_2=bilby.core.prior.Sine(name='tilt_2', boundary='reflective'),
                phi_12=bilby.core.prior.Uniform(
                    name='phi_12', minimum=0, maximum=2 * np.pi, boundary='periodic'),
                phi_jl=bilby.core.prior.Uniform(
                    name='phi_jl', minimum=0, maximum=2 * np.pi, boundary='periodic'),
                luminosity_distance=bilby.gw.prior.UniformSourceFrame(
                    name='luminosity_distance', minimum=1e2,
                    maximum=5e3, unit='Mpc', boundary=None),
                dec=bilby.core.prior.Cosine(name='dec', boundary='reflective'),
                ra=bilby.core.prior.Uniform(
                    name='ra', minimum=0, maximum=2 * np.pi, boundary='periodic'),
                theta_jn=bilby.core.prior.Sine(name='theta_jn', boundary='reflective'),
                psi=bilby.core.prior.Uniform(name='psi', minimum=0, maximum=np.pi, boundary='periodic'),
                phase=bilby.core.prior.Uniform(
                    name='phase', minimum=0, maximum=2 * np.pi, boundary='periodic')
            )
        )
        all_keys = set(prior_set.keys()).union(set(expected.keys()))
        for key in all_keys:
            self.assertEqual(expected[key], prior_set[key])

    def test_dict_argument_is_not_string_or_dict(self):
        with self.assertRaises(ValueError):
            bilby.core.prior.PriorDict(dictionary=list())

    def test_sample_subset_correct_size(self):
        size = 7
        samples = self.prior_set_from_dict.sample_subset(keys=self.prior_set_from_dict.keys(), size=size)
        self.assertEqual(len(self.prior_set_from_dict), len(samples))
        for key in samples:
            self.assertEqual(size, len(samples[key]))

    def test_sample_subset_correct_size_when_non_priors_in_dict(self):
        self.prior_set_from_dict['asdf'] = 'not_a_prior'
        samples = self.prior_set_from_dict.sample_subset(keys=self.prior_set_from_dict.keys())
        self.assertEqual(len(self.prior_set_from_dict) - 1, len(samples))

    def test_sample_subset_with_actual_subset(self):
        size = 3
        samples = self.prior_set_from_dict.sample_subset(keys=['length'], size=size)
        expected = dict(length=np.array([42., 42., 42.]))
        self.assertTrue(np.array_equal(expected['length'], samples['length']))

    def test_sample_subset_constrained_as_array(self):
        size = 3
        keys = ["mass", "speed"]
        out = self.prior_set_from_dict.sample_subset_constrained_as_array(keys, size)
        self.assertTrue(isinstance(out, np.ndarray))
        self.assertTrue(out.shape == (len(keys), size))

    def test_sample(self):
        size = 7
        np.random.seed(42)
        samples1 = self.prior_set_from_dict.sample_subset(keys=self.prior_set_from_dict.keys(), size=size)
        np.random.seed(42)
        samples2 = self.prior_set_from_dict.sample(size=size)
        self.assertEqual(set(samples1.keys()), set(samples2.keys()))
        for key in samples1:
            self.assertTrue(np.array_equal(samples1[key], samples2[key]))

    def test_prob(self):
        samples = self.prior_set_from_dict.sample_subset(keys=['mass', 'speed'])
        expected = self.first_prior.prob(samples['mass']) * self.second_prior.prob(samples['speed'])
        self.assertEqual(expected, self.prior_set_from_dict.prob(samples))

    def test_ln_prob(self):
        samples = self.prior_set_from_dict.sample_subset(keys=['mass', 'speed'])
        expected = self.first_prior.ln_prob(samples['mass']) + self.second_prior.ln_prob(samples['speed'])
        self.assertEqual(expected, self.prior_set_from_dict.ln_prob(samples))

    def test_rescale(self):
        theta = [0.5, 0.5, 0.5]
        expected = [self.first_prior.rescale(0.5),
                    self.second_prior.rescale(0.5),
                    self.third_prior.rescale(0.5)]
        self.assertListEqual(sorted(expected), sorted(self.prior_set_from_dict.rescale(
            keys=self.prior_set_from_dict.keys(), theta=theta)))

    def test_redundancy(self):
        for key in self.prior_set_from_dict.keys():
            self.assertFalse(self.prior_set_from_dict.test_redundancy(key=key))


class TestConstraintPriorNormalisation(unittest.TestCase):
    def setUp(self):
        self.priors = dict(mass_1=bilby.core.prior.Uniform(name='mass_1', minimum=5, maximum=10, unit='$M_{\odot}$',
                                                           boundary=None),
                      mass_2=bilby.core.prior.Uniform(name='mass_2', minimum=5, maximum=10, unit='$M_{\odot}$',
                                                      boundary=None),
                      mass_ratio=bilby.core.prior.Constraint(name='mass_ratio', minimum=0, maximum=1))
        self.priors = bilby.core.prior.PriorDict(self.priors)

    def test_prob_integrate_to_one(self):
        keys = ['mass_1', 'mass_2', 'mass_ratio']
        n = 5000
        samples = self.priors.sample_subset(keys=keys, size=n)
        prob = self.priors.prob(samples, axis=0)
        dm1 = self.priors['mass_1'].maximum - self.priors['mass_1'].minimum
        dm2 = self.priors['mass_2'].maximum - self.priors['mass_2'].minimum
        integral = np.sum(prob * (dm1 * dm2)) / len(samples['mass_1'])
        self.assertAlmostEqual(1, integral, 5)


class TestLoadPrior(unittest.TestCase):
    def test_load_prior_with_float(self):
        filename = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                                'prior_files/prior_with_floats.prior')
        prior = bilby.core.prior.PriorDict(filename)
        self.assertTrue("mass_1" in prior)
        self.assertTrue("mass_2" in prior)
        self.assertTrue(prior['mass_2'].peak == 20)

    def test_load_prior_with_parentheses(self):
        filename = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                                'prior_files/prior_with_parentheses.prior')
        prior = bilby.core.prior.PriorDict(filename)
        self.assertTrue(isinstance(prior['logA'], bilby.core.prior.Uniform))


class TestFillPrior(unittest.TestCase):

    def setUp(self):
        self.likelihood = Mock()
        self.likelihood.parameters = dict(a=0, b=0, c=0, d=0, asdf=0, ra=1)
        self.likelihood.non_standard_sampling_parameter_keys = dict(t=8)
        self.priors = dict(a=1, b=1.1, c="string", d=bilby.core.prior.Uniform(0, 1))
        self.priors = bilby.core.prior.PriorDict(dictionary=self.priors)
        self.default_prior_file = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                                               'prior_files/binary_black_holes.prior')
        self.priors.fill_priors(self.likelihood, self.default_prior_file)

    def tearDown(self):
        del self.likelihood
        del self.priors

    def test_prior_instances_are_not_changed_by_parsing(self):
        self.assertIsInstance(self.priors['d'], bilby.core.prior.Uniform)

    def test_parsing_ints_to_delta_priors_class(self):
        self.assertIsInstance(self.priors['a'], bilby.core.prior.DeltaFunction)

    def test_parsing_ints_to_delta_priors_with_right_value(self):
        self.assertEqual(self.priors['a'].peak, 1)

    def test_parsing_floats_to_delta_priors_class(self):
        self.assertIsInstance(self.priors['b'], bilby.core.prior.DeltaFunction)

    def test_parsing_floats_to_delta_priors_with_right_value(self):
        self.assertAlmostEqual(self.priors['b'].peak, 1.1, 1e-8)

    def test_without_available_default_priors_no_prior_is_set(self):
        with self.assertRaises(KeyError):
            print(self.priors['asdf'])

    def test_with_available_default_priors_a_default_prior_is_set(self):
        self.assertIsInstance(self.priors['ra'], bilby.core.prior.Uniform)


class TestCreateDefaultPrior(unittest.TestCase):

    def test_none_behaviour(self):
        self.assertIsNone(bilby.core.prior.create_default_prior(name='name', default_priors_file=None))

    def test_bbh_params(self):
        prior_file = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'prior_files/binary_black_holes.prior')
        prior_set = bilby.core.prior.PriorDict(filename=prior_file)
        for prior in prior_set:
            self.assertEqual(prior_set[prior], bilby.core.prior.create_default_prior(name=prior,
                                                                                     default_priors_file=prior_file))

    def test_unknown_prior(self):
        prior_file = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'prior_files/binary_black_holes.prior')
        self.assertIsNone(bilby.core.prior.create_default_prior(name='name', default_priors_file=prior_file))


class TestConditionalPrior(unittest.TestCase):

    def setUp(self):
        self.condition_func_call_counter = 0

        def condition_func(reference_parameters, test_variable_1, test_variable_2):
            self.condition_func_call_counter += 1
            return {key: value + 1 for key, value in reference_parameters.items()}
        self.condition_func = condition_func
        self.minimum = 0
        self.maximum = 5
        self.test_variable_1 = 0
        self.test_variable_2 = 1
        self.prior = bilby.core.prior.ConditionalBasePrior(condition_func=condition_func,
                                                           minimum=self.minimum,
                                                           maximum=self.maximum)

    def tearDown(self):
        del self.condition_func
        del self.condition_func_call_counter
        del self.minimum
        del self.maximum
        del self.test_variable_1
        del self.test_variable_2
        del self.prior

    def test_reference_params(self):
        self.assertDictEqual(dict(minimum=self.minimum, maximum=self.maximum), self.prior.reference_params)

    def test_required_variables(self):
        self.assertListEqual(['test_variable_1', 'test_variable_2'], sorted(self.prior.required_variables))

    def test_required_variables_no_condition_func(self):
        self.prior = bilby.core.prior.ConditionalBasePrior(condition_func=None,
                                                           minimum=self.minimum,
                                                           maximum=self.maximum)
        self.assertListEqual([], self.prior.required_variables)

    def test_get_instantiation_dict(self):
        expected = dict(minimum=0, maximum=5, name=None, latex_label=None, unit=None,
                        boundary=None, condition_func=self.condition_func)
        actual = self.prior.get_instantiation_dict()
        for key, value in expected.items():
            if key == 'condition_func':
                continue
            self.assertEqual(value, actual[key])

    def test_update_conditions_correct_variables(self):
        self.prior.update_conditions(test_variable_1=self.test_variable_1, test_variable_2=self.test_variable_2)
        self.assertEqual(1, self.condition_func_call_counter)
        self.assertEqual(self.minimum + 1, self.prior.minimum)
        self.assertEqual(self.maximum + 1, self.prior.maximum)

    def test_update_conditions_no_variables(self):
        self.prior.update_conditions(test_variable_1=self.test_variable_1, test_variable_2=self.test_variable_2)
        self.prior.update_conditions()
        self.assertEqual(1, self.condition_func_call_counter)
        self.assertEqual(self.minimum + 1, self.prior.minimum)
        self.assertEqual(self.maximum + 1, self.prior.maximum)

    def test_update_conditions_illegal_variables(self):
        with self.assertRaises(bilby.core.prior.IllegalRequiredVariablesException):
            self.prior.update_conditions(test_parameter_1=self.test_variable_1)

    def test_sample_calls_update_conditions(self):
        with mock.patch.object(self.prior, 'update_conditions') as m:
            self.prior.sample(1,
                              test_parameter_1=self.test_variable_1,
                              test_parameter_2=self.test_variable_2)
            m.assert_called_with(test_parameter_1=self.test_variable_1, test_parameter_2=self.test_variable_2)

    def test_rescale_calls_update_conditions(self):
        with mock.patch.object(self.prior, 'update_conditions') as m:
            self.prior.rescale(1, test_parameter_1=self.test_variable_1,
                               test_parameter_2=self.test_variable_2)
            m.assert_called_with(test_parameter_1=self.test_variable_1,
                                 test_parameter_2=self.test_variable_2)

    def test_rescale_prob_update_conditions(self):
        with mock.patch.object(self.prior, 'update_conditions') as m:
            self.prior.prob(1, test_parameter_1=self.test_variable_1,
                            test_parameter_2=self.test_variable_2)
            m.assert_called_with(test_parameter_1=self.test_variable_1,
                                 test_parameter_2=self.test_variable_2)

    def test_rescale_ln_prob_update_conditions(self):
        with mock.patch.object(self.prior, 'update_conditions') as m:
            self.prior.ln_prob(1, test_parameter_1=self.test_variable_1,
                               test_parameter_2=self.test_variable_2)
            calls = [mock.call(test_parameter_1=self.test_variable_1,
                               test_parameter_2=self.test_variable_2),
                     mock.call()]
            m.assert_has_calls(calls)

    def test_reset_to_reference_parameters(self):
        self.prior.minimum = 10
        self.prior.maximum = 20
        self.prior.reset_to_reference_parameters()
        self.assertEqual(self.prior.reference_params['minimum'], self.prior.minimum)
        self.assertEqual(self.prior.reference_params['maximum'], self.prior.maximum)

    def test_cond_prior_instantiation_no_boundary_prior(self):
        prior = bilby.core.prior.ConditionalFermiDirac(condition_func=None, sigma=1, mu=1)
        self.assertIsNone(prior.boundary)


class TestConditionalPriorDict(unittest.TestCase):

    def setUp(self):
        def condition_func_1(reference_parameters, var_0):
            return reference_parameters

        def condition_func_2(reference_parameters, var_0, var_1):
            return reference_parameters

        def condition_func_3(reference_parameters, var_1, var_2):
            return reference_parameters

        self.minimum = 0
        self.maximum = 1
        self.prior_0 = bilby.core.prior.Uniform(minimum=self.minimum, maximum=self.maximum)
        self.prior_1 = bilby.core.prior.ConditionalUniform(condition_func=condition_func_1,
                                                           minimum=self.minimum, maximum=self.maximum)
        self.prior_2 = bilby.core.prior.ConditionalUniform(condition_func=condition_func_2,
                                                           minimum=self.minimum, maximum=self.maximum)
        self.prior_3 = bilby.core.prior.ConditionalUniform(condition_func=condition_func_3,
                                                           minimum=self.minimum, maximum=self.maximum)
        self.conditional_priors = bilby.core.prior.ConditionalPriorDict(dict(var_3=self.prior_3, var_2=self.prior_2,
                                                                             var_0=self.prior_0, var_1=self.prior_1))
        self.conditional_priors_manually_set_items = bilby.core.prior.ConditionalPriorDict()
        self.test_sample = dict(var_0=0.3, var_1=0.4, var_2=0.5, var_3=0.4)
        for key, value in dict(var_0=self.prior_0, var_1=self.prior_1, var_2=self.prior_2, var_3=self.prior_3).items():
            self.conditional_priors_manually_set_items[key] = value

    def tearDown(self):
        del self.minimum
        del self.maximum
        del self.prior_0
        del self.prior_1
        del self.prior_2
        del self.prior_3
        del self.conditional_priors
        del self.conditional_priors_manually_set_items
        del self.test_sample

    def test_conditions_resolved_upon_instantiation(self):
        self.assertListEqual(['var_0', 'var_1', 'var_2', 'var_3'], self.conditional_priors.sorted_keys)

    def test_conditions_resolved_setting_items(self):
        self.assertListEqual(['var_0', 'var_1', 'var_2', 'var_3'],
                             self.conditional_priors_manually_set_items.sorted_keys)

    def test_unconditional_keys_upon_instantiation(self):
        self.assertListEqual(['var_0'], self.conditional_priors.unconditional_keys)

    def test_unconditional_keys_setting_items(self):
        self.assertListEqual(['var_0'], self.conditional_priors_manually_set_items.unconditional_keys)

    def test_conditional_keys_upon_instantiation(self):
        self.assertListEqual(['var_1', 'var_2', 'var_3'], self.conditional_priors.conditional_keys)

    def test_conditional_keys_setting_items(self):
        self.assertListEqual(['var_1', 'var_2', 'var_3'], self.conditional_priors_manually_set_items.conditional_keys)

    def test_prob(self):
        self.assertEqual(1, self.conditional_priors.prob(sample=self.test_sample))

    def test_prob_illegal_conditions(self):
        del self.conditional_priors['var_0']
        with self.assertRaises(bilby.core.prior.IllegalConditionsException):
            self.conditional_priors.prob(sample=self.test_sample)

    def test_ln_prob(self):
        self.assertEqual(0, self.conditional_priors.ln_prob(sample=self.test_sample))

    def test_ln_prob_illegal_conditions(self):
        del self.conditional_priors['var_0']
        with self.assertRaises(bilby.core.prior.IllegalConditionsException):
            self.conditional_priors.ln_prob(sample=self.test_sample)

    def test_sample_subset_all_keys(self):
        with mock.patch("numpy.random.uniform") as m:
            m.return_value = 0.5
            self.assertDictEqual(dict(var_0=0.5, var_1=0.5, var_2=0.5, var_3=0.5),
                                 self.conditional_priors.sample_subset(keys=['var_0', 'var_1', 'var_2', 'var_3']))

    def test_sample_illegal_subset(self):
        with mock.patch("numpy.random.uniform") as m:
            m.return_value = 0.5
            with self.assertRaises(bilby.core.prior.IllegalConditionsException):
                self.conditional_priors.sample_subset(keys=['var_1'])

    def test_sample_multiple(self):
        def condition_func(reference_params, a):
            return dict(minimum=reference_params['minimum'],
                        maximum=reference_params['maximum'],
                        alpha=reference_params['alpha'] * a)
        priors = bilby.core.prior.ConditionalPriorDict()
        priors['a'] = bilby.core.prior.Uniform(minimum=0, maximum=1)
        priors['b'] = bilby.core.prior.ConditionalPowerLaw(condition_func=condition_func, minimum=1, maximum=2,
                                                           alpha=-2)
        print(priors.sample(2))

    def test_rescale(self):

        def condition_func_1_rescale(reference_parameters, var_0):
            if var_0 == 0.5:
                return dict(minimum=reference_parameters['minimum'], maximum=1)
            return reference_parameters

        def condition_func_2_rescale(reference_parameters, var_0, var_1):
            if var_0 == 0.5 and var_1 == 0.5:
                return dict(minimum=reference_parameters['minimum'], maximum=1)
            return reference_parameters

        def condition_func_3_rescale(reference_parameters, var_1, var_2):
            if var_1 == 0.5 and var_2 == 0.5:
                return dict(minimum=reference_parameters['minimum'], maximum=1)
            return reference_parameters

        self.prior_0 = bilby.core.prior.Uniform(minimum=self.minimum, maximum=1)
        self.prior_1 = bilby.core.prior.ConditionalUniform(condition_func=condition_func_1_rescale,
                                                           minimum=self.minimum, maximum=2)
        self.prior_2 = bilby.core.prior.ConditionalUniform(condition_func=condition_func_2_rescale,
                                                           minimum=self.minimum, maximum=2)
        self.prior_3 = bilby.core.prior.ConditionalUniform(condition_func=condition_func_3_rescale,
                                                           minimum=self.minimum, maximum=2)
        self.conditional_priors = bilby.core.prior.ConditionalPriorDict(dict(var_3=self.prior_3, var_2=self.prior_2,
                                                                             var_0=self.prior_0, var_1=self.prior_1))
        ref_variables = [0.5, 0.5, 0.5, 0.5]
        res = self.conditional_priors.rescale(keys=list(self.test_sample.keys()),
                                              theta=ref_variables)
        self.assertListEqual(ref_variables, res)

    def test_rescale_illegal_conditions(self):
        del self.conditional_priors['var_0']
        with self.assertRaises(bilby.core.prior.IllegalConditionsException):
            self.conditional_priors.rescale(keys=list(self.test_sample.keys()), theta=list(self.test_sample.values()))

    def test_combined_conditions(self):
        def d_condition_func(reference_params, a, b, c):
            return dict(minimum=reference_params['minimum'], maximum=reference_params['maximum'])

        def a_condition_func(reference_params, b, c):
            return dict(minimum=reference_params['minimum'], maximum=reference_params['maximum'])

        priors = bilby.core.prior.ConditionalPriorDict()

        priors['a'] = bilby.core.prior.ConditionalUniform(condition_func=a_condition_func,
                                                          minimum=0, maximum=1)

        priors['b'] = bilby.core.prior.LogUniform(minimum=1, maximum=10)

        priors['d'] = bilby.core.prior.ConditionalUniform(condition_func=d_condition_func,
                                                          minimum=0.0, maximum=1.0)

        priors['c'] = bilby.core.prior.LogUniform(minimum=1, maximum=10)
        sample = priors.sample()
        res = priors.rescale(['a', 'b', 'd', 'c'], [0.5, 0.5, 0.5, 0.5])
        print(res)


class TestJsonIO(unittest.TestCase):

    def setUp(self):
        mvg = bilby.core.prior.MultivariateGaussianDist(names=['testa', 'testb'],
                                                        mus=[1, 1],
                                                        covs=np.array([[2., 0.5], [0.5, 2.]]),
                                                        weights=1.)
        mvn = bilby.core.prior.MultivariateGaussianDist(names=['testa', 'testb'],
                                                        mus=[1, 1],
                                                        covs=np.array([[2., 0.5], [0.5, 2.]]),
                                                        weights=1.)
        hp_map_file = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                                   'prior_files/GW150914_testing_skymap.fits')
        hp_dist = bilby.gw.prior.HealPixMapPriorDist(hp_map_file,
                                                 names=['testra', 'testdec'])
        hp_3d_dist = bilby.gw.prior.HealPixMapPriorDist(hp_map_file,
                                                        names=['testra', 'testdec', 'testdistance'], distance=True)

        self.priors = bilby.core.prior.PriorDict(dict(
            a=bilby.core.prior.DeltaFunction(name='test', unit='unit', peak=1),
            b=bilby.core.prior.Gaussian(name='test', unit='unit', mu=0, sigma=1),
            c=bilby.core.prior.Normal(name='test', unit='unit', mu=0, sigma=1),
            d=bilby.core.prior.PowerLaw(name='test', unit='unit', alpha=0, minimum=0, maximum=1),
            e=bilby.core.prior.PowerLaw(name='test', unit='unit', alpha=-1, minimum=0.5, maximum=1),
            f=bilby.core.prior.PowerLaw(name='test', unit='unit', alpha=2, minimum=1, maximum=1e2),
            g=bilby.core.prior.Uniform(name='test', unit='unit', minimum=0, maximum=1),
            h=bilby.core.prior.LogUniform(name='test', unit='unit', minimum=5e0, maximum=1e2),
            i=bilby.gw.prior.UniformComovingVolume(name='redshift', minimum=0.1, maximum=1.0),
            j=bilby.gw.prior.UniformSourceFrame(name='luminosity_distance', minimum=1.0, maximum=1000.0),
            k=bilby.core.prior.Sine(name='test', unit='unit'),
            l=bilby.core.prior.Cosine(name='test', unit='unit'),
            m=bilby.core.prior.Interped(name='test', unit='unit', xx=np.linspace(0, 10, 1000),
                                        yy=np.linspace(0, 10, 1000) ** 4,
                                        minimum=3, maximum=5),
            n=bilby.core.prior.TruncatedGaussian(name='test', unit='unit', mu=1, sigma=0.4, minimum=-1, maximum=1),
            o=bilby.core.prior.TruncatedNormal(name='test', unit='unit', mu=1, sigma=0.4, minimum=-1, maximum=1),
            p=bilby.core.prior.HalfGaussian(name='test', unit='unit', sigma=1),
            q=bilby.core.prior.HalfNormal(name='test', unit='unit', sigma=1),
            r=bilby.core.prior.LogGaussian(name='test', unit='unit', mu=0, sigma=1),
            s=bilby.core.prior.LogNormal(name='test', unit='unit', mu=0, sigma=1),
            t=bilby.core.prior.Exponential(name='test', unit='unit', mu=1),
            u=bilby.core.prior.StudentT(name='test', unit='unit', df=3, mu=0, scale=1),
            v=bilby.core.prior.Beta(name='test', unit='unit', alpha=2.0, beta=2.0),
            x=bilby.core.prior.Logistic(name='test', unit='unit', mu=0, scale=1),
            y=bilby.core.prior.Cauchy(name='test', unit='unit', alpha=0, beta=1),
            z=bilby.core.prior.Lorentzian(name='test', unit='unit', alpha=0, beta=1),
            aa=bilby.core.prior.Gamma(name='test', unit='unit', k=1, theta=1),
            ab=bilby.core.prior.ChiSquared(name='test', unit='unit', nu=2),
            ac=bilby.gw.prior.AlignedSpin(name='test', unit='unit'),
            ad=bilby.core.prior.MultivariateGaussian(dist=mvg, name='testa', unit='unit'),
            ae=bilby.core.prior.MultivariateGaussian(dist=mvg, name='testb', unit='unit'),
            af=bilby.core.prior.MultivariateNormal(dist=mvn, name='testa', unit='unit'),
            ag=bilby.core.prior.MultivariateNormal(dist=mvn, name='testb', unit='unit'),
            ah=bilby.gw.prior.HealPixPrior(dist=hp_dist, name='testra', unit='unit'),
            ai=bilby.gw.prior.HealPixPrior(dist=hp_dist, name='testdec', unit='unit'),
            aj=bilby.gw.prior.HealPixPrior(dist=hp_3d_dist, name='testra', unit='unit'),
            ak=bilby.gw.prior.HealPixPrior(dist=hp_3d_dist, name='testdec', unit='unit'),
            al=bilby.gw.prior.HealPixPrior(dist=hp_3d_dist, name='testdistance', unit='unit')
        ))

    def test_read_write_to_json(self):
        """ Interped prior is removed as there is numerical error in the recovered prior."""
        self.priors.to_json(outdir="prior_files", label="json_test")
        new_priors = bilby.core.prior.PriorDict.from_json(filename="prior_files/json_test_prior.json")
        old_interped = self.priors.pop("m")
        new_interped = new_priors.pop("m")
        self.assertDictEqual(self.priors, new_priors)
        self.assertLess(max(abs(old_interped.xx - new_interped.xx)), 1e-15)
        self.assertLess(max(abs(old_interped.yy - new_interped.yy)), 1e-15)


if __name__ == '__main__':
    unittest.main()



from __future__ import absolute_import
import shutil
import os
import logging
from packaging import version

import unittest
import numpy as np
import bilby
import scipy
from scipy.stats import ks_2samp, kstest


def ks_2samp_wrapper(data1, data2):
    if version.parse(scipy.__version__) >= version.parse("1.3.0"):
        return ks_2samp(data1, data2, alternative="two-sided", mode="asymp")
    else:
        return ks_2samp(data1, data2)


class Test(unittest.TestCase):
    outdir = 'outdir_for_tests'

    @classmethod
    def setUpClass(self):
        if os.path.isdir(self.outdir):
            try:
                shutil.rmtree(self.outdir)
            except OSError:
                logging.warning(
                    "{} not removed prior to tests".format(self.outdir))

    @classmethod
    def tearDownClass(self):
        if os.path.isdir(self.outdir):
            try:
                shutil.rmtree(self.outdir)
            except OSError:
                logging.warning(
                    "{} not removed prior to tests".format(self.outdir))

    def test_fifteen_dimensional_cbc(self):
        duration = 4.
        sampling_frequency = 2048.
        label = 'full_15_parameters'
        np.random.seed(88170235)

        waveform_arguments = dict(waveform_approximant='IMRPhenomPv2',
                                  reference_frequency=50., minimum_frequency=20.)
        waveform_generator = bilby.gw.WaveformGenerator(
            duration=duration, sampling_frequency=sampling_frequency,
            frequency_domain_source_model=bilby.gw.source.lal_binary_black_hole,
            parameter_conversion=bilby.gw.conversion.convert_to_lal_binary_black_hole_parameters,
            waveform_arguments=waveform_arguments)

        ifos = bilby.gw.detector.InterferometerList(['H1', 'L1'])
        ifos.set_strain_data_from_power_spectral_densities(
            sampling_frequency=sampling_frequency, duration=duration,
            start_time=0)

        priors = bilby.gw.prior.BBHPriorDict()
        priors.pop('mass_1')
        priors.pop('mass_2')
        priors['chirp_mass'] = bilby.prior.Uniform(
            name='chirp_mass', latex_label='$M$', minimum=10.0, maximum=100.0,
            unit='$M_{\\odot}$')
        priors['mass_ratio'] = bilby.prior.Uniform(
            name='mass_ratio', latex_label='$q$', minimum=0.5, maximum=1.0)
        priors['geocent_time'] = bilby.core.prior.Uniform(
            minimum=-0.1, maximum=0.1)

        likelihood = bilby.gw.GravitationalWaveTransient(
            interferometers=ifos, waveform_generator=waveform_generator,
            priors=priors, distance_marginalization=False,
            phase_marginalization=False, time_marginalization=False)

        likelihood = bilby.core.likelihood.ZeroLikelihood(likelihood)

        result = bilby.run_sampler(
            likelihood=likelihood, priors=priors, sampler='dynesty',
            npoints=1000, walks=100, outdir=self.outdir, label=label)
        pvalues = [ks_2samp_wrapper(result.priors[key].sample(10000),
                                    result.posterior[key].values).pvalue
                   for key in priors.keys()]
        print("P values per parameter")
        for key, p in zip(priors.keys(), pvalues):
            print(key, p)
        self.assertGreater(kstest(pvalues, "uniform").pvalue, 0.01)


if __name__ == '__main__':
    unittest.main()



from __future__ import division, absolute_import
import unittest
import os

import numpy as np
import bilby
from bilby.gw.likelihood import BilbyROQParamsRangeError


class TestBasicGWTransient(unittest.TestCase):

    def setUp(self):
        np.random.seed(500)
        self.parameters = dict(
            mass_1=31., mass_2=29., a_1=0.4, a_2=0.3, tilt_1=0.0, tilt_2=0.0,
            phi_12=1.7, phi_jl=0.3, luminosity_distance=4000., theta_jn=0.4,
            psi=2.659, phase=1.3, geocent_time=1126259642.413, ra=1.375,
            dec=-1.2108)
        self.interferometers = bilby.gw.detector.InterferometerList(['H1'])
        self.interferometers.set_strain_data_from_power_spectral_densities(
            sampling_frequency=2048, duration=4)
        self.waveform_generator = bilby.gw.waveform_generator.WaveformGenerator(
            duration=4, sampling_frequency=2048,
            frequency_domain_source_model=bilby.gw.source.lal_binary_black_hole,
            )

        self.likelihood = bilby.gw.likelihood.BasicGravitationalWaveTransient(
            interferometers=self.interferometers,
            waveform_generator=self.waveform_generator
        )
        self.likelihood.parameters = self.parameters.copy()

    def tearDown(self):
        del self.parameters
        del self.interferometers
        del self.waveform_generator
        del self.likelihood

    def test_noise_log_likelihood(self):
        """Test noise log likelihood matches precomputed value"""
        self.likelihood.noise_log_likelihood()
        self.assertAlmostEqual(-4037.0994372143414, self.likelihood.noise_log_likelihood(), 3)

    def test_log_likelihood(self):
        """Test log likelihood matches precomputed value"""
        self.likelihood.log_likelihood()
        self.assertAlmostEqual(self.likelihood.log_likelihood(),
                               -4055.265194950631, 3)

    def test_log_likelihood_ratio(self):
        """Test log likelihood ratio returns the correct value"""
        self.assertAlmostEqual(
            self.likelihood.log_likelihood()
            - self.likelihood.noise_log_likelihood(),
            self.likelihood.log_likelihood_ratio(), 3)

    def test_likelihood_zero_when_waveform_is_none(self):
        """Test log likelihood returns np.nan_to_num(-np.inf) when the
        waveform is None"""
        self.likelihood.waveform_generator.frequency_domain_strain = \
            lambda x: None
        self.assertEqual(self.likelihood.log_likelihood_ratio(),
                         np.nan_to_num(-np.inf))

    def test_repr(self):
        expected = 'BasicGravitationalWaveTransient(interferometers={},\n\twaveform_generator={})'.format(
            self.interferometers, self.waveform_generator)
        self.assertEqual(expected, repr(self.likelihood))


class TestGWTransient(unittest.TestCase):

    def setUp(self):
        np.random.seed(500)
        self.duration = 4
        self.sampling_frequency = 2048
        self.parameters = dict(
            mass_1=31., mass_2=29., a_1=0.4, a_2=0.3, tilt_1=0.0, tilt_2=0.0,
            phi_12=1.7, phi_jl=0.3, luminosity_distance=4000., theta_jn=0.4,
            psi=2.659, phase=1.3, geocent_time=1126259642.413, ra=1.375,
            dec=-1.2108)
        self.interferometers = bilby.gw.detector.InterferometerList(['H1'])
        self.interferometers.set_strain_data_from_power_spectral_densities(
            sampling_frequency=self.sampling_frequency, duration=self.duration)
        self.waveform_generator = bilby.gw.waveform_generator.WaveformGenerator(
            duration=self.duration, sampling_frequency=self.sampling_frequency,
            frequency_domain_source_model=bilby.gw.source.lal_binary_black_hole,
        )

        self.prior = bilby.gw.prior.BBHPriorDict()
        self.prior['geocent_time'] = bilby.prior.Uniform(
            minimum=self.parameters['geocent_time'] - self.duration / 2,
            maximum=self.parameters['geocent_time'] + self.duration / 2)

        self.likelihood = bilby.gw.likelihood.GravitationalWaveTransient(
            interferometers=self.interferometers,
            waveform_generator=self.waveform_generator, priors=self.prior.copy()
        )
        self.likelihood.parameters = self.parameters.copy()

    def tearDown(self):
        del self.parameters
        del self.interferometers
        del self.waveform_generator
        del self.prior
        del self.likelihood

    def test_noise_log_likelihood(self):
        """Test noise log likelihood matches precomputed value"""
        self.likelihood.noise_log_likelihood()
        self.assertAlmostEqual(-4037.0994372143414, self.likelihood.noise_log_likelihood(), 3)

    def test_log_likelihood(self):
        """Test log likelihood matches precomputed value"""
        self.likelihood.log_likelihood()
        self.assertAlmostEqual(self.likelihood.log_likelihood(),
                               -4055.265194950631, 3)

    def test_log_likelihood_ratio(self):
        """Test log likelihood ratio returns the correct value"""
        self.assertAlmostEqual(
            self.likelihood.log_likelihood()
            - self.likelihood.noise_log_likelihood(),
            self.likelihood.log_likelihood_ratio(), 3)

    def test_likelihood_zero_when_waveform_is_none(self):
        """Test log likelihood returns np.nan_to_num(-np.inf) when the
        waveform is None"""
        self.likelihood.waveform_generator.frequency_domain_strain =\
            lambda x: None
        self.assertEqual(self.likelihood.log_likelihood_ratio(),
                         np.nan_to_num(-np.inf))

    def test_repr(self):
        expected = 'GravitationalWaveTransient(interferometers={},\n\twaveform_generator={},\n\t' \
                   'time_marginalization={}, distance_marginalization={}, phase_marginalization={}, ' \
                   'priors={})'.format(self.interferometers, self.waveform_generator, False, False, False, self.prior)
        self.assertEqual(expected, repr(self.likelihood))

    def test_interferometers_setting_list(self):
        ifos = [bilby.gw.detector.get_empty_interferometer(name=name) for name in ['H1', 'L1']]
        self.likelihood.interferometers = ifos
        self.assertListEqual(bilby.gw.detector.InterferometerList(ifos), self.likelihood.interferometers)
        self.assertTrue(type(self.likelihood.interferometers) == bilby.gw.detector.InterferometerList)

    def test_interferometers_setting_interferometer_list(self):
        ifos = bilby.gw.detector.InterferometerList([bilby.gw.detector.get_empty_interferometer(name=name)
                                                     for name in ['H1', 'L1']])
        self.likelihood.interferometers = ifos
        self.assertListEqual(bilby.gw.detector.InterferometerList(ifos), self.likelihood.interferometers)
        self.assertTrue(type(self.likelihood.interferometers) == bilby.gw.detector.InterferometerList)

    def test_meta_data(self):
        expected = dict(
            interferometers=self.interferometers.meta_data,
            time_marginalization=False,
            phase_marginalization=False,
            distance_marginalization=False,
            waveform_generator_class=self.waveform_generator.__class__,
            waveform_arguments=self.waveform_generator.waveform_arguments,
            frequency_domain_source_model=self.waveform_generator.frequency_domain_source_model,
            parameter_conversion=self.waveform_generator.parameter_conversion,
            sampling_frequency=self.waveform_generator.sampling_frequency,
            duration=self.waveform_generator.duration,
            start_time=self.waveform_generator.start_time,
            lal_version=self.likelihood.lal_version)
        self.assertDictEqual(expected, self.likelihood.meta_data)


class TestTimeMarginalization(unittest.TestCase):

    def setUp(self):
        np.random.seed(500)
        self.duration = 4
        self.sampling_frequency = 2048
        self.parameters = dict(
            mass_1=31., mass_2=29., a_1=0.4, a_2=0.3, tilt_1=0.0, tilt_2=0.0,
            phi_12=1.7, phi_jl=0.3, luminosity_distance=4000., theta_jn=0.4,
            psi=2.659, phase=1.3, geocent_time=1126259640, ra=1.375,
            dec=-1.2108)

        self.interferometers = bilby.gw.detector.InterferometerList(['H1'])
        self.interferometers.set_strain_data_from_power_spectral_densities(
            sampling_frequency=self.sampling_frequency, duration=self.duration,
            start_time=1126259640)

        self.waveform_generator = bilby.gw.waveform_generator.WaveformGenerator(
            duration=self.duration, sampling_frequency=self.sampling_frequency,
            frequency_domain_source_model=bilby.gw.source.lal_binary_black_hole,
            start_time=1126259640)

        self.priors = bilby.gw.prior.BBHPriorDict()

        self.likelihood = bilby.gw.likelihood.GravitationalWaveTransient(
            interferometers=self.interferometers,
            waveform_generator=self.waveform_generator, priors=self.priors.copy()
        )

        self.likelihood.parameters = self.parameters.copy()

    def tearDown(self):
        del self.duration
        del self.sampling_frequency
        del self.parameters
        del self.interferometers
        del self.waveform_generator
        del self.priors
        del self.likelihood

    def test_time_marginalisation_full_segment(self):
        """
        Test time marginalised likelihood matches brute force version over the
        whole segment.
        """
        likes = []
        lls = []
        self.priors['geocent_time'] = bilby.prior.Uniform(
            minimum=self.waveform_generator.start_time,
            maximum=self.waveform_generator.start_time + self.duration)
        self.time = bilby.gw.likelihood.GravitationalWaveTransient(
            interferometers=self.interferometers,
            waveform_generator=self.waveform_generator,
            time_marginalization=True, priors=self.priors
        )
        times = self.waveform_generator.start_time + np.linspace(
            0, self.duration, 4097)[:-1]
        for time in times:
            self.likelihood.parameters['geocent_time'] = time
            lls.append(self.likelihood.log_likelihood_ratio())
            likes.append(np.exp(lls[-1]))

        marg_like = np.log(np.trapz(
            likes * self.time.priors['geocent_time'].prob(times), times))
        self.time.parameters = self.parameters.copy()
        self.time.parameters['time_jitter'] = 0.0
        self.time.parameters['geocent_time'] = self.waveform_generator.start_time
        self.assertAlmostEqual(marg_like, self.time.log_likelihood_ratio(),
                               delta=0.5)

    def test_time_marginalisation_partial_segment(self):
        """
        Test time marginalised likelihood matches brute force version over the
        whole segment.
        """
        likes = []
        lls = []
        self.priors['geocent_time'] = bilby.prior.Uniform(
            minimum=self.parameters['geocent_time'] + 1 - 0.1,
            maximum=self.parameters['geocent_time'] + 1 + 0.1)
        self.time = bilby.gw.likelihood.GravitationalWaveTransient(
            interferometers=self.interferometers,
            waveform_generator=self.waveform_generator,
            time_marginalization=True, priors=self.priors
        )
        times = self.waveform_generator.start_time + np.linspace(
            0, self.duration, 4097)[:-1]
        for time in times:
            self.likelihood.parameters['geocent_time'] = time
            lls.append(self.likelihood.log_likelihood_ratio())
            likes.append(np.exp(lls[-1]))

        marg_like = np.log(np.trapz(
            likes * self.time.priors['geocent_time'].prob(times), times))
        self.time.parameters = self.parameters.copy()
        self.time.parameters['time_jitter'] = 0.0
        self.time.parameters['geocent_time'] = self.waveform_generator.start_time
        self.assertAlmostEqual(marg_like, self.time.log_likelihood_ratio(),
                               delta=0.5)


class TestMarginalizedLikelihood(unittest.TestCase):

    def setUp(self):
        np.random.seed(500)
        self.duration = 4
        self.sampling_frequency = 2048
        self.parameters = dict(
            mass_1=31., mass_2=29., a_1=0.4, a_2=0.3, tilt_1=0.0, tilt_2=0.0,
            phi_12=1.7, phi_jl=0.3, luminosity_distance=4000., theta_jn=0.4,
            psi=2.659, phase=1.3, geocent_time=1126259642.413, ra=1.375,
            dec=-1.2108)

        self.interferometers = bilby.gw.detector.InterferometerList(['H1'])
        self.interferometers.set_strain_data_from_power_spectral_densities(
            sampling_frequency=self.sampling_frequency, duration=self.duration,
            start_time=self.parameters['geocent_time'] - self.duration / 2)

        self.waveform_generator = bilby.gw.waveform_generator.WaveformGenerator(
            duration=self.duration, sampling_frequency=self.sampling_frequency,
            frequency_domain_source_model=bilby.gw.source.lal_binary_black_hole,
        )

        self.prior = bilby.gw.prior.BBHPriorDict()
        self.prior['geocent_time'] = bilby.prior.Uniform(
            minimum=self.parameters['geocent_time'] - self.duration / 2,
            maximum=self.parameters['geocent_time'] + self.duration / 2)

    def test_cannot_instantiate_marginalised_likelihood_without_prior(self):
        self.assertRaises(
            ValueError,
            lambda: bilby.gw.likelihood.GravitationalWaveTransient(
                interferometers=self.interferometers,
                waveform_generator=self.waveform_generator,
                phase_marginalization=True))

    def test_generating_default_time_prior(self):
        temp = self.prior.pop('geocent_time')
        new_prior = self.prior.copy()
        like = bilby.gw.likelihood.GravitationalWaveTransient(
            interferometers=self.interferometers,
            waveform_generator=self.waveform_generator, priors=new_prior,
            time_marginalization=True
        )
        same = all([temp.minimum == like.priors['geocent_time'].minimum,
                    temp.maximum == like.priors['geocent_time'].maximum,
                    new_prior['geocent_time'] == temp.minimum])
        self.assertTrue(same)
        self.prior['geocent_time'] = temp

    def test_generating_default_phase_prior(self):
        temp = self.prior.pop('phase')
        new_prior = self.prior.copy()
        like = bilby.gw.likelihood.GravitationalWaveTransient(
            interferometers=self.interferometers,
            waveform_generator=self.waveform_generator, priors=new_prior,
            phase_marginalization=True
        )
        same = all([temp.minimum == like.priors['phase'].minimum,
                    temp.maximum == like.priors['phase'].maximum,
                    new_prior['phase'] == float(0)])
        self.assertTrue(same)
        self.prior['phase'] = temp

    def test_run_sampler_flags_if_marginalized_phase_is_sampled(self):
        like = bilby.gw.likelihood.GravitationalWaveTransient(
            interferometers=self.interferometers,
            waveform_generator=self.waveform_generator, priors=self.prior,
            phase_marginalization=True
        )
        new_prior = self.prior.copy()
        new_prior['phase'] = bilby.prior.Uniform(minimum=0, maximum=2*np.pi)
        for key, param in dict(
            mass_1=31., mass_2=29., a_1=0.4, a_2=0.3, tilt_1=0.0, tilt_2=0.0,
            phi_12=1.7, phi_jl=0.3, theta_jn=0.4, psi=2.659, ra=1.375, dec=-1.2108).items():
            new_prior[key] = param
        with self.assertRaises(bilby.core.sampler.SamplingMarginalisedParameterError):
            bilby.run_sampler(like, new_prior)

    def test_run_sampler_flags_if_marginalized_time_is_sampled(self):
        like = bilby.gw.likelihood.GravitationalWaveTransient(
            interferometers=self.interferometers,
            waveform_generator=self.waveform_generator, priors=self.prior,
            time_marginalization=True
        )
        new_prior = self.prior.copy()
        new_prior['geocent_time'] = bilby.prior.Uniform(minimum=0, maximum=1)
        for key, param in dict(
            mass_1=31., mass_2=29., a_1=0.4, a_2=0.3, tilt_1=0.0, tilt_2=0.0,
            phi_12=1.7, phi_jl=0.3, theta_jn=0.4, psi=2.659, ra=1.375, dec=-1.2108).items():
            new_prior[key] = param
        with self.assertRaises(bilby.core.sampler.SamplingMarginalisedParameterError):
            bilby.run_sampler(like, new_prior)


class TestPhaseMarginalization(unittest.TestCase):

    def setUp(self):
        np.random.seed(500)
        self.duration = 4
        self.sampling_frequency = 2048
        self.parameters = dict(
            mass_1=31., mass_2=29., a_1=0.4, a_2=0.3, tilt_1=0.0, tilt_2=0.0,
            phi_12=1.7, phi_jl=0.3, luminosity_distance=4000., theta_jn=0.4,
            psi=2.659, phase=1.3, geocent_time=1126259642.413, ra=1.375,
            dec=-1.2108)

        self.interferometers = bilby.gw.detector.InterferometerList(['H1'])
        self.interferometers.set_strain_data_from_power_spectral_densities(
            sampling_frequency=self.sampling_frequency, duration=self.duration)

        self.waveform_generator = bilby.gw.waveform_generator.WaveformGenerator(
            duration=self.duration, sampling_frequency=self.sampling_frequency,
            frequency_domain_source_model=bilby.gw.source.lal_binary_black_hole,
        )

        self.prior = bilby.gw.prior.BBHPriorDict()
        self.prior['geocent_time'] = bilby.prior.Uniform(
            minimum=self.parameters['geocent_time'] - self.duration / 2,
            maximum=self.parameters['geocent_time'] + self.duration / 2)

        self.likelihood = bilby.gw.likelihood.GravitationalWaveTransient(
            interferometers=self.interferometers,
            waveform_generator=self.waveform_generator, priors=self.prior.copy()
        )

        self.phase = bilby.gw.likelihood.GravitationalWaveTransient(
            interferometers=self.interferometers,
            waveform_generator=self.waveform_generator,
            phase_marginalization=True, priors=self.prior.copy()
        )
        for like in [self.likelihood, self.phase]:
            like.parameters = self.parameters.copy()

    def tearDown(self):
        del self.duration
        del self.sampling_frequency
        del self.parameters
        del self.interferometers
        del self.waveform_generator
        del self.prior
        del self.likelihood
        del self.phase

    def test_phase_marginalisation(self):
        """Test phase marginalised likelihood matches brute force version"""
        like = []
        phases = np.linspace(0, 2 * np.pi, 1000)
        for phase in phases:
            self.likelihood.parameters['phase'] = phase
            like.append(np.exp(self.likelihood.log_likelihood_ratio()))

        marg_like = np.log(np.trapz(like, phases) / (2 * np.pi))
        self.phase.parameters = self.parameters.copy()
        self.assertAlmostEqual(marg_like, self.phase.log_likelihood_ratio(),
                               delta=0.5)


class TestTimePhaseMarginalization(unittest.TestCase):

    def setUp(self):
        np.random.seed(500)
        self.duration = 4
        self.sampling_frequency = 2048
        self.parameters = dict(
            mass_1=31., mass_2=29., a_1=0.4, a_2=0.3, tilt_1=0.0, tilt_2=0.0,
            phi_12=1.7, phi_jl=0.3, luminosity_distance=4000., theta_jn=0.4,
            psi=2.659, phase=1.3, geocent_time=1126259642.413, ra=1.375,
            dec=-1.2108)

        self.interferometers = bilby.gw.detector.InterferometerList(['H1'])
        self.interferometers.set_strain_data_from_power_spectral_densities(
            sampling_frequency=self.sampling_frequency, duration=self.duration,
            start_time=1126259640)

        self.waveform_generator = bilby.gw.waveform_generator.WaveformGenerator(
            duration=self.duration, sampling_frequency=self.sampling_frequency,
            frequency_domain_source_model=bilby.gw.source.lal_binary_black_hole,
            start_time=1126259640)

        self.priors = bilby.gw.prior.BBHPriorDict()
        self.priors['geocent_time'] = bilby.prior.Uniform(
            minimum=self.parameters['geocent_time'] - self.duration / 2,
            maximum=self.parameters['geocent_time'] + self.duration / 2)

        self.likelihood = bilby.gw.likelihood.GravitationalWaveTransient(
            interferometers=self.interferometers,
            waveform_generator=self.waveform_generator, priors=self.priors.copy()
        )

        self.time = bilby.gw.likelihood.GravitationalWaveTransient(
            interferometers=self.interferometers,
            waveform_generator=self.waveform_generator,
            time_marginalization=True, priors=self.priors.copy()
        )

        self.phase = bilby.gw.likelihood.GravitationalWaveTransient(
            interferometers=self.interferometers,
            waveform_generator=self.waveform_generator,
            phase_marginalization=True, priors=self.priors.copy()
        )

        self.time_phase = bilby.gw.likelihood.GravitationalWaveTransient(
            interferometers=self.interferometers,
            waveform_generator=self.waveform_generator,
            time_marginalization=True, phase_marginalization=True,
            priors=self.priors
        )
        for like in [self.likelihood, self.time, self.phase, self.time_phase]:
            like.parameters = self.parameters.copy()

    def tearDown(self):
        del self.duration
        del self.sampling_frequency
        del self.parameters
        del self.interferometers
        del self.waveform_generator
        del self.priors
        del self.likelihood
        del self.time
        del self.phase
        del self.time_phase

    def test_time_marginalisation(self):
        """
        Test time marginalised likelihood matches brute force version when
        also marginalising over phase.
        """
        like = []
        times = np.linspace(
            self.time_phase.priors['geocent_time'].minimum,
            self.time_phase.priors['geocent_time'].maximum, 4097)[:-1]
        for time in times:
            self.phase.parameters['geocent_time'] = time
            like.append(np.exp(self.phase.log_likelihood_ratio()))

        marg_like = np.log(np.trapz(like, times)
                           / self.waveform_generator.duration)
        self.time_phase.parameters = self.parameters.copy()
        self.time_phase.parameters['time_jitter'] = 0.0
        self.assertAlmostEqual(marg_like,
                               self.time_phase.log_likelihood_ratio(),
                               delta=0.5)

    def test_phase_marginalisation(self):
        """
        Test phase marginalised likelihood matches brute force version when
        also marginalising over time.
        """
        like = []
        phases = np.linspace(0, 2 * np.pi, 1000)
        for phase in phases:
            self.time.parameters['phase'] = phase
            self.time.parameters['time_jitter'] = 0.0
            like.append(np.exp(self.time.log_likelihood_ratio()))

        marg_like = np.log(np.trapz(like, phases) / (2 * np.pi))
        self.time_phase.parameters = self.parameters.copy()
        self.time_phase.parameters['time_jitter'] = 0.0
        self.assertAlmostEqual(marg_like,
                               self.time_phase.log_likelihood_ratio(),
                               delta=0.5)


class TestROQLikelihood(unittest.TestCase):

    def setUp(self):
        self.duration = 4
        self.sampling_frequency = 2048

        # Possible locations for the ROQ: in the docker image, local, or on CIT
        trial_roq_paths = [
            "/roq_basis",
            os.path.join(os.path.expanduser("~"), 'ROQ_data/IMRPhenomPv2/4s'),
            "/home/cbc/ROQ_data/IMRPhenomPv2/4s"]
        roq_dir = None
        for path in trial_roq_paths:
            if os.path.isdir(path):
                roq_dir = path
                break
        if roq_dir is None:
            raise Exception("Unable to load ROQ basis: cannot proceed with tests")

        linear_matrix_file = "{}/B_linear.npy".format(roq_dir)
        quadratic_matrix_file = "{}/B_quadratic.npy".format(roq_dir)

        fnodes_linear_file = "{}/fnodes_linear.npy".format(roq_dir)
        fnodes_linear = np.load(fnodes_linear_file).T
        fnodes_quadratic_file = "{}/fnodes_quadratic.npy".format(roq_dir)
        fnodes_quadratic = np.load(fnodes_quadratic_file).T
        self.linear_matrix_file = "{}/B_linear.npy".format(roq_dir)
        self.quadratic_matrix_file = "{}/B_quadratic.npy".format(roq_dir)
        self.params_file = "{}/params.dat".format(roq_dir)

        self.test_parameters = dict(
            mass_1=36.0, mass_2=36.0, a_1=0.0, a_2=0.0, tilt_1=0.0, tilt_2=0.0,
            phi_12=1.7, phi_jl=0.3, luminosity_distance=1000., theta_jn=0.4,
            psi=0.659, phase=1.3, geocent_time=1.2, ra=1.3, dec=-1.2)

        ifos = bilby.gw.detector.InterferometerList(['H1'])
        ifos.set_strain_data_from_power_spectral_densities(
            sampling_frequency=self.sampling_frequency, duration=self.duration)

        self.priors = bilby.gw.prior.BBHPriorDict()
        self.priors.pop("mass_1")
        self.priors.pop("mass_2")
        # Testing is done with the 4s IMRPhenomPV2 ROQ basis
        self.priors["chirp_mass"] = bilby.core.prior.Uniform(12.299703, 45)
        self.priors["mass_ratio"] = bilby.core.prior.Uniform(0.125, 1)
        self.priors['geocent_time'] = bilby.core.prior.Uniform(1.19, 1.21)

        non_roq_wfg = bilby.gw.WaveformGenerator(
            duration=self.duration, sampling_frequency=self.sampling_frequency,
            frequency_domain_source_model=bilby.gw.source.lal_binary_black_hole,
            waveform_arguments=dict(
                reference_frequency=20.0, minimum_frequency=20.0,
                approximant='IMRPhenomPv2'))

        ifos.inject_signal(
            parameters=self.test_parameters, waveform_generator=non_roq_wfg)

        self.ifos = ifos

        roq_wfg = bilby.gw.waveform_generator.WaveformGenerator(
            duration=self.duration, sampling_frequency=self.sampling_frequency,
            frequency_domain_source_model=bilby.gw.source.roq,
            waveform_arguments=dict(
                frequency_nodes_linear=fnodes_linear,
                frequency_nodes_quadratic=fnodes_quadratic,
                reference_frequency=20., minimum_frequency=20.,
                approximant='IMRPhenomPv2'))

        self.roq_wfg = roq_wfg

        self.non_roq = bilby.gw.likelihood.GravitationalWaveTransient(
            interferometers=ifos, waveform_generator=non_roq_wfg)

        self.non_roq_phase = bilby.gw.likelihood.GravitationalWaveTransient(
            interferometers=ifos, waveform_generator=non_roq_wfg,
            phase_marginalization=True, priors=self.priors.copy())

        self.roq = bilby.gw.likelihood.ROQGravitationalWaveTransient(
            interferometers=ifos, waveform_generator=roq_wfg,
            linear_matrix=linear_matrix_file,
            quadratic_matrix=quadratic_matrix_file, priors=self.priors)

        self.roq_phase = bilby.gw.likelihood.ROQGravitationalWaveTransient(
            interferometers=ifos, waveform_generator=roq_wfg,
            linear_matrix=linear_matrix_file,
            quadratic_matrix=quadratic_matrix_file,
            phase_marginalization=True, priors=self.priors.copy())

    def tearDown(self):
        del (self.roq, self.non_roq, self.non_roq_phase, self.roq_phase,
             self.ifos, self.priors)

    def test_matches_non_roq(self):
        self.non_roq.parameters.update(self.test_parameters)
        self.roq.parameters.update(self.test_parameters)
        self.assertLess(
            abs(self.non_roq.log_likelihood_ratio() -
                self.roq.log_likelihood_ratio()) /
            self.non_roq.log_likelihood_ratio(), 1e-3)

    def test_time_prior_out_of_bounds_returns_zero(self):
        self.roq.parameters.update(self.test_parameters)
        self.roq.parameters['geocent_time'] = -5
        self.assertEqual(
            self.roq.log_likelihood_ratio(), np.nan_to_num(-np.inf))

    def test_phase_marginalisation_roq(self):
        """Test phase marginalised likelihood matches brute force version"""
        self.non_roq_phase.parameters = self.test_parameters.copy()
        self.roq_phase.parameters = self.test_parameters.copy()
        self.assertLess(
            abs(self.non_roq_phase.log_likelihood_ratio() -
                self.roq_phase.log_likelihood_ratio()) /
            self.non_roq_phase.log_likelihood_ratio(), 1e-3)

    def test_create_roq_weights_with_params(self):
        roq = bilby.gw.likelihood.ROQGravitationalWaveTransient(
            interferometers=self.ifos, waveform_generator=self.roq_wfg,
            linear_matrix=self.linear_matrix_file, roq_params=self.params_file,
            quadratic_matrix=self.quadratic_matrix_file, priors=self.priors)
        roq.parameters.update(self.test_parameters)
        self.roq.parameters.update(self.test_parameters)
        self.assertEqual(
            roq.log_likelihood_ratio(), self.roq.log_likelihood_ratio())

    def test_create_roq_weights_frequency_mismatch_works_with_params(self):

        self.ifos[0].maximum_frequency = self.ifos[0].maximum_frequency / 2
        bilby.gw.likelihood.ROQGravitationalWaveTransient(
            interferometers=self.ifos, waveform_generator=self.roq_wfg,
            linear_matrix=self.linear_matrix_file, roq_params=self.params_file,
            quadratic_matrix=self.quadratic_matrix_file, priors=self.priors)

    def test_create_roq_weights_frequency_mismatch_fails_without_params(self):
        self.ifos[0].maximum_frequency = self.ifos[0].maximum_frequency / 2
        with self.assertRaises(ValueError):
            bilby.gw.likelihood.ROQGravitationalWaveTransient(
                interferometers=self.ifos, waveform_generator=self.roq_wfg,
                linear_matrix=self.linear_matrix_file,
                quadratic_matrix=self.quadratic_matrix_file, priors=self.priors)

    def test_create_roq_weights_fails_with_min_chirp_mass_outside_bounds(self):
        self.ifos[0].maximum_frequency = self.ifos[0].maximum_frequency / 2
        self.priors["chirp_mass"] = bilby.core.prior.Uniform(10, 45)
        with self.assertRaises(BilbyROQParamsRangeError):
            bilby.gw.likelihood.ROQGravitationalWaveTransient(
                interferometers=self.ifos, waveform_generator=self.roq_wfg,
                linear_matrix=self.linear_matrix_file,
                roq_params=self.params_file,
                quadratic_matrix=self.quadratic_matrix_file,
                priors=self.priors)

    def test_create_roq_weights_fails_with_max_chirp_mass_outside_bounds(self):
        self.ifos[0].maximum_frequency = self.ifos[0].maximum_frequency / 2
        self.priors["chirp_mass"] = bilby.core.prior.Uniform(12.299703, 50)
        with self.assertRaises(BilbyROQParamsRangeError):
            bilby.gw.likelihood.ROQGravitationalWaveTransient(
                interferometers=self.ifos, waveform_generator=self.roq_wfg,
                linear_matrix=self.linear_matrix_file,
                roq_params=self.params_file,
                quadratic_matrix=self.quadratic_matrix_file,
                priors=self.priors)

    def test_create_roq_weights_fails_with_min_component_mass_outside_bounds(self):
        self.ifos[0].maximum_frequency = self.ifos[0].maximum_frequency / 2
        self.priors["chirp_mass"] = bilby.core.prior.Uniform(12.299703, 45)
        self.priors["mass_ratio"] = bilby.core.prior.Uniform(1e-5, 1)
        with self.assertRaises(BilbyROQParamsRangeError):
            bilby.gw.likelihood.ROQGravitationalWaveTransient(
                interferometers=self.ifos, waveform_generator=self.roq_wfg,
                linear_matrix=self.linear_matrix_file,
                roq_params=self.params_file,
                quadratic_matrix=self.quadratic_matrix_file,
                priors=self.priors)

    def test_create_roq_weights_fails_with_max_frequency(self):
        ifos = bilby.gw.detector.InterferometerList(['H1'])
        ifos.set_strain_data_from_power_spectral_densities(
            sampling_frequency=2**14, duration=4)
        ifos[0].maximum_frequency = 2**13
        with self.assertRaises(BilbyROQParamsRangeError):
            bilby.gw.likelihood.ROQGravitationalWaveTransient(
                interferometers=ifos, waveform_generator=self.roq_wfg,
                linear_matrix=self.linear_matrix_file,
                roq_params=self.params_file,
                quadratic_matrix=self.quadratic_matrix_file,
                priors=self.priors)

    def test_create_roq_weights_fails_due_to_min_frequency(self):
        self.ifos[0].minimum_frequency = 15
        with self.assertRaises(BilbyROQParamsRangeError):
            bilby.gw.likelihood.ROQGravitationalWaveTransient(
                interferometers=self.ifos, waveform_generator=self.roq_wfg,
                linear_matrix=self.linear_matrix_file,
                roq_params=self.params_file,
                quadratic_matrix=self.quadratic_matrix_file,
                priors=self.priors)

    def test_create_roq_weights_fails_due_to_duration(self):
        ifos = bilby.gw.detector.InterferometerList(['H1'])
        ifos.set_strain_data_from_power_spectral_densities(
            sampling_frequency=self.sampling_frequency, duration=16)
        with self.assertRaises(BilbyROQParamsRangeError):
            bilby.gw.likelihood.ROQGravitationalWaveTransient(
                interferometers=ifos, waveform_generator=self.roq_wfg,
                linear_matrix=self.linear_matrix_file,
                roq_params=self.params_file,
                quadratic_matrix=self.quadratic_matrix_file,
                priors=self.priors)


class TestRescaledROQLikelihood(unittest.TestCase):

    def test_rescaling(self):

        # Possible locations for the ROQ: in the docker image, local, or on CIT
        trial_roq_paths = [
            "/roq_basis",
            os.path.join(os.path.expanduser("~"), 'ROQ_data/IMRPhenomPv2/4s'),
            "/home/cbc/ROQ_data/IMRPhenomPv2/4s"]
        roq_dir = None
        for path in trial_roq_paths:
            if os.path.isdir(path):
                roq_dir = path
                break
        if roq_dir is None:
            raise Exception("Unable to load ROQ basis: cannot proceed with tests")

        linear_matrix_file = "{}/B_linear.npy".format(roq_dir)
        quadratic_matrix_file = "{}/B_quadratic.npy".format(roq_dir)

        fnodes_linear_file = "{}/fnodes_linear.npy".format(roq_dir)
        fnodes_linear = np.load(fnodes_linear_file).T
        fnodes_quadratic_file = "{}/fnodes_quadratic.npy".format(roq_dir)
        fnodes_quadratic = np.load(fnodes_quadratic_file).T
        self.linear_matrix_file = "{}/B_linear.npy".format(roq_dir)
        self.quadratic_matrix_file = "{}/B_quadratic.npy".format(roq_dir)
        self.params_file = "{}/params.dat".format(roq_dir)

        scale_factor = 0.5
        params = np.genfromtxt(self.params_file, names=True)

        self.duration = 4 / scale_factor
        self.sampling_frequency = 2048 * scale_factor

        ifos = bilby.gw.detector.InterferometerList(['H1'])
        ifos.set_strain_data_from_power_spectral_densities(
            sampling_frequency=self.sampling_frequency, duration=self.duration)
        self.ifos = ifos

        self.priors = bilby.gw.prior.BBHPriorDict()
        self.priors.pop("mass_1")
        self.priors.pop("mass_2")
        # Testing is done with the 4s IMRPhenomPV2 ROQ basis
        self.priors["chirp_mass"] = bilby.core.prior.Uniform(
            12.299703 / scale_factor, 45 / scale_factor)
        self.priors["mass_ratio"] = bilby.core.prior.Uniform(0.125, 1)
        self.priors['geocent_time'] = bilby.core.prior.Uniform(1.19, 1.21)

        self.roq_wfg = bilby.gw.waveform_generator.WaveformGenerator(
            duration=self.duration, sampling_frequency=self.sampling_frequency,
            frequency_domain_source_model=bilby.gw.source.roq,
            waveform_arguments=dict(
                frequency_nodes_linear=fnodes_linear,
                frequency_nodes_quadratic=fnodes_quadratic,
                reference_frequency=20., minimum_frequency=20.,
                approximant='IMRPhenomPv2'))

        self.roq = bilby.gw.likelihood.ROQGravitationalWaveTransient(
            interferometers=ifos, waveform_generator=self.roq_wfg,
            linear_matrix=linear_matrix_file, roq_params=params, 
            roq_scale_factor=scale_factor, quadratic_matrix=quadratic_matrix_file,
            priors=self.priors)


class TestBBHLikelihoodSetUp(unittest.TestCase):

    def setUp(self):
        self.ifos = bilby.gw.detector.InterferometerList(['H1'])

    def tearDown(self):
        del self.ifos

    def test_instantiation(self):
        self.like = bilby.gw.likelihood.get_binary_black_hole_likelihood(
            self.ifos)


if __name__ == '__main__':
    unittest.main()



import os
import shutil
import unittest

import pandas as pd

import bilby


class TestCBCResult(unittest.TestCase):

    def setUp(self):
        bilby.utils.command_line_args.bilby_test_mode = False
        priors = bilby.gw.prior.BBHPriorDict()
        priors['geocent_time'] = 2
        injection_parameters = priors.sample()
        self.meta_data = dict(
            likelihood=dict(
                phase_marginalization=True,
                distance_marginalization=False,
                time_marginalization=True,
                frequency_domain_source_model=bilby.gw.source.lal_binary_black_hole,
                waveform_arguments=dict(
                    reference_frequency=20.0,
                    waveform_approximant='IMRPhenomPv2'),
                interferometers=dict(
                    H1=dict(optimal_SNR=1, parameters=injection_parameters),
                    L1=dict(optimal_SNR=1, parameters=injection_parameters)),
                sampling_frequency=4096,
                duration=4,
                start_time=0,
                waveform_generator_class=bilby.gw.waveform_generator.WaveformGenerator,
                parameter_conversion=bilby.gw.conversion.convert_to_lal_binary_black_hole_parameters,
            )
        )
        self.result = bilby.gw.result.CBCResult(
            label='label', outdir='outdir', sampler='nestle',
            search_parameter_keys=list(priors.keys()), fixed_parameter_keys=list(),
            priors=priors, sampler_kwargs=dict(test='test', func=lambda x: x),
            injection_parameters=injection_parameters,
            meta_data=self.meta_data,
            posterior=pd.DataFrame(priors.sample(100))
        )
        if not os.path.isdir(self.result.outdir):
            os.mkdir(self.result.outdir)
        pass

    def tearDown(self):
        bilby.utils.command_line_args.bilby_test_mode = True
        try:
            shutil.rmtree(self.result.outdir)
        except OSError:
            pass
        del self.result
        pass

    def test_phase_marginalization(self):
        self.assertEqual(
            self.result.phase_marginalization,
            self.meta_data['likelihood']['phase_marginalization'])

    def test_phase_marginalization_unset(self):
        self.result.meta_data['likelihood'].pop('phase_marginalization')
        with self.assertRaises(AttributeError):
            self.result.phase_marginalization

    def test_time_marginalization(self):
        self.assertEqual(
            self.result.time_marginalization,
            self.meta_data['likelihood']['time_marginalization'])

    def test_time_marginalization_unset(self):
        self.result.meta_data['likelihood'].pop('time_marginalization')
        with self.assertRaises(AttributeError):
            self.result.time_marginalization

    def test_distance_marginalization(self):
        self.assertEqual(
            self.result.distance_marginalization,
            self.meta_data['likelihood']['distance_marginalization'])

    def test_distance_marginalization_unset(self):
        self.result.meta_data['likelihood'].pop('distance_marginalization')
        with self.assertRaises(AttributeError):
            self.result.distance_marginalization

    def test_reference_frequency(self):
        self.assertEqual(
            self.result.reference_frequency,
            self.meta_data['likelihood']['waveform_arguments']['reference_frequency'])

    def test_reference_frequency_unset(self):
        self.result.meta_data['likelihood']['waveform_arguments'].pop('reference_frequency')
        with self.assertRaises(AttributeError):
            self.result.reference_frequency

    def test_sampling_frequency(self):
        self.assertEqual(
            self.result.sampling_frequency,
            self.meta_data['likelihood']['sampling_frequency'])

    def test_sampling_frequency_unset(self):
        self.result.meta_data['likelihood'].pop('sampling_frequency')
        with self.assertRaises(AttributeError):
            self.result.sampling_frequency

    def test_duration(self):
        self.assertEqual(
            self.result.duration,
            self.meta_data['likelihood']['duration'])

    def test_duration_unset(self):
        self.result.meta_data['likelihood'].pop('duration')
        with self.assertRaises(AttributeError):
            self.result.duration

    def test_start_time(self):
        self.assertEqual(
            self.result.start_time,
            self.meta_data['likelihood']['start_time'])

    def test_start_time_unset(self):
        self.result.meta_data['likelihood'].pop('start_time')
        with self.assertRaises(AttributeError):
            self.result.start_time

    def test_waveform_approximant(self):
        self.assertEqual(
            self.result.waveform_approximant,
            self.meta_data['likelihood']['waveform_arguments']['waveform_approximant'])

    def test_waveform_approximant_unset(self):
        self.result.meta_data['likelihood']['waveform_arguments'].pop('waveform_approximant')
        with self.assertRaises(AttributeError):
            self.result.waveform_approximant

    def test_waveform_arguments(self):
        self.assertEqual(
            self.result.waveform_arguments,
            self.meta_data['likelihood']['waveform_arguments'])

    def test_frequency_domain_source_model(self):
        self.assertEqual(
            self.result.frequency_domain_source_model,
            self.meta_data['likelihood']['frequency_domain_source_model'])

    def test_frequency_domain_source_model_unset(self):
        self.result.meta_data['likelihood'].pop('frequency_domain_source_model')
        with self.assertRaises(AttributeError):
            self.result.frequency_domain_source_model

    def test_parameter_conversion(self):
        self.assertEqual(
            self.result.parameter_conversion,
            self.meta_data['likelihood']['parameter_conversion'])

    def test_parameter_conversion_unset(self):
        self.result.meta_data['likelihood'].pop('parameter_conversion')
        with self.assertRaises(AttributeError):
            self.result.parameter_conversion

    def test_waveform_generator_class(self):
        self.assertEqual(
            self.result.waveform_generator_class,
            self.meta_data['likelihood']['waveform_generator_class'])

    def test_waveform_generator_class_unset(self):
        self.result.meta_data['likelihood'].pop('waveform_generator_class')
        with self.assertRaises(AttributeError):
            self.result.waveform_generator_class

    def test_interferometer_names(self):
        self.assertEqual(
            self.result.interferometers,
            [name for name in self.meta_data['likelihood']['interferometers']])

    def test_detector_injection_properties(self):
        self.assertEqual(
            self.result.detector_injection_properties('H1'),
            self.meta_data['likelihood']['interferometers']['H1'])

    def test_detector_injection_properties_no_injection(self):
        self.assertEqual(
            self.result.detector_injection_properties('not_a_detector'),
            None)


if __name__ == '__main__':
    unittest.main()
